from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.schema_property_data_type import SchemaPropertyDataType
from ..models.schema_property_format import SchemaPropertyFormat
from ..models.schema_property_type import SchemaPropertyType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.schema_property_items import SchemaPropertyItems
    from ..models.survey_answer import SurveyAnswer


T = TypeVar("T", bound="SchemaProperty")


@_attrs_define
class SchemaProperty:
    """Each entry represents a field in the `values` field of the JSON survey response

    Attributes:
        type (SchemaPropertyType):
        description (str):
        format_ (Union[Unset, SchemaPropertyFormat]):
        question_id (Union[Unset, str]): The ID of the question this field is associated with
        loop_number (Union[Unset, int]): For loop and merge questions, this field documents which loop number the field
            in the JSON instance represents
        export_tag (Union[Unset, str]): The export tag of this field
        question_export_tag (Union[Unset, str]): The export tag of the question this field is associated with
        data_type (Union[Unset, SchemaPropertyDataType]): The type of data this field represents
        one_of (Union[Unset, List['SurveyAnswer']]):
        items (Union[Unset, SchemaPropertyItems]):
    """

    type: SchemaPropertyType
    description: str
    format_: Union[Unset, SchemaPropertyFormat] = UNSET
    question_id: Union[Unset, str] = UNSET
    loop_number: Union[Unset, int] = UNSET
    export_tag: Union[Unset, str] = UNSET
    question_export_tag: Union[Unset, str] = UNSET
    data_type: Union[Unset, SchemaPropertyDataType] = UNSET
    one_of: Union[Unset, List["SurveyAnswer"]] = UNSET
    items: Union[Unset, "SchemaPropertyItems"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type.value

        description = self.description
        format_: Union[Unset, str] = UNSET
        if not isinstance(self.format_, Unset):
            format_ = self.format_.value

        question_id = self.question_id
        loop_number = self.loop_number
        export_tag = self.export_tag
        question_export_tag = self.question_export_tag
        data_type: Union[Unset, str] = UNSET
        if not isinstance(self.data_type, Unset):
            data_type = self.data_type.value

        one_of: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.one_of, Unset):
            one_of = []
            for one_of_item_data in self.one_of:
                one_of_item = one_of_item_data.to_dict()

                one_of.append(one_of_item)

        items: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.items, Unset):
            items = self.items.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "description": description,
            }
        )
        if format_ is not UNSET:
            field_dict["format"] = format_
        if question_id is not UNSET:
            field_dict["questionId"] = question_id
        if loop_number is not UNSET:
            field_dict["loopNumber"] = loop_number
        if export_tag is not UNSET:
            field_dict["exportTag"] = export_tag
        if question_export_tag is not UNSET:
            field_dict["questionExportTag"] = question_export_tag
        if data_type is not UNSET:
            field_dict["dataType"] = data_type
        if one_of is not UNSET:
            field_dict["oneOf"] = one_of
        if items is not UNSET:
            field_dict["items"] = items

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.schema_property_items import SchemaPropertyItems
        from ..models.survey_answer import SurveyAnswer

        d = src_dict.copy()
        type = SchemaPropertyType(d.pop("type"))

        description = d.pop("description")

        _format_ = d.pop("format", UNSET)
        format_: Union[Unset, SchemaPropertyFormat]
        if isinstance(_format_, Unset):
            format_ = UNSET
        else:
            format_ = SchemaPropertyFormat(_format_)

        question_id = d.pop("questionId", UNSET)

        loop_number = d.pop("loopNumber", UNSET)

        export_tag = d.pop("exportTag", UNSET)

        question_export_tag = d.pop("questionExportTag", UNSET)

        _data_type = d.pop("dataType", UNSET)
        data_type: Union[Unset, SchemaPropertyDataType]
        if isinstance(_data_type, Unset):
            data_type = UNSET
        else:
            data_type = SchemaPropertyDataType(_data_type)

        one_of = []
        _one_of = d.pop("oneOf", UNSET)
        for one_of_item_data in _one_of or []:
            one_of_item = SurveyAnswer.from_dict(one_of_item_data)

            one_of.append(one_of_item)

        _items = d.pop("items", UNSET)
        items: Union[Unset, SchemaPropertyItems]
        if isinstance(_items, Unset):
            items = UNSET
        else:
            items = SchemaPropertyItems.from_dict(_items)

        schema_property = cls(
            type=type,
            description=description,
            format_=format_,
            question_id=question_id,
            loop_number=loop_number,
            export_tag=export_tag,
            question_export_tag=question_export_tag,
            data_type=data_type,
            one_of=one_of,
            items=items,
        )

        schema_property.additional_properties = d
        return schema_property

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .create_response_request import CreateResponseRequest
from .create_response_request_values import CreateResponseRequestValues
from .create_response_response import CreateResponseResponse
from .create_response_response_result import CreateResponseResponseResult
from .create_response_with_file_attachments_request import (
    CreateResponseWithFileAttachmentsRequest,
)
from .create_response_with_file_attachments_request_file_mapping import (
    CreateResponseWithFileAttachmentsRequestFileMapping,
)
from .default_error_response import DefaultErrorResponse
from .delete_response_response import DeleteResponseResponse
from .error import Error
from .get_schema_response import GetSchemaResponse
from .get_schema_response_result import GetSchemaResponseResult
from .get_schema_response_result_properties import GetSchemaResponseResultProperties
from .get_schema_response_result_properties_displayed_fields import (
    GetSchemaResponseResultPropertiesDisplayedFields,
)
from .get_schema_response_result_properties_displayed_fields_items import (
    GetSchemaResponseResultPropertiesDisplayedFieldsItems,
)
from .get_schema_response_result_properties_displayed_fields_items_type import (
    GetSchemaResponseResultPropertiesDisplayedFieldsItemsType,
)
from .get_schema_response_result_properties_displayed_fields_type import (
    GetSchemaResponseResultPropertiesDisplayedFieldsType,
)
from .get_schema_response_result_properties_displayed_values import (
    GetSchemaResponseResultPropertiesDisplayedValues,
)
from .get_schema_response_result_properties_displayed_values_additional_properties import (
    GetSchemaResponseResultPropertiesDisplayedValuesAdditionalProperties,
)
from .get_schema_response_result_properties_displayed_values_additional_properties_type import (
    GetSchemaResponseResultPropertiesDisplayedValuesAdditionalPropertiesType,
)
from .get_schema_response_result_properties_displayed_values_type import (
    GetSchemaResponseResultPropertiesDisplayedValuesType,
)
from .get_schema_response_result_properties_labels import (
    GetSchemaResponseResultPropertiesLabels,
)
from .get_schema_response_result_properties_labels_additional_properties import (
    GetSchemaResponseResultPropertiesLabelsAdditionalProperties,
)
from .get_schema_response_result_properties_labels_type import (
    GetSchemaResponseResultPropertiesLabelsType,
)
from .get_schema_response_result_properties_response_id import (
    GetSchemaResponseResultPropertiesResponseId,
)
from .get_schema_response_result_properties_response_id_type import (
    GetSchemaResponseResultPropertiesResponseIdType,
)
from .get_schema_response_result_properties_values import (
    GetSchemaResponseResultPropertiesValues,
)
from .get_schema_response_result_properties_values_properties import (
    GetSchemaResponseResultPropertiesValuesProperties,
)
from .get_schema_response_result_properties_values_type import (
    GetSchemaResponseResultPropertiesValuesType,
)
from .get_schema_response_result_schema import GetSchemaResponseResultSchema
from .get_schema_response_result_type import GetSchemaResponseResultType
from .meta import Meta
from .meta_with_error import MetaWithError
from .retrieve_response_response import RetrieveResponseResponse
from .retrieve_response_response_not_yet_available import (
    RetrieveResponseResponseNotYetAvailable,
)
from .retrieve_response_response_not_yet_available_result import (
    RetrieveResponseResponseNotYetAvailableResult,
)
from .retrieve_response_response_result import RetrieveResponseResponseResult
from .retrieve_response_response_result_displayed_values import (
    RetrieveResponseResponseResultDisplayedValues,
)
from .retrieve_response_response_result_labels import (
    RetrieveResponseResponseResultLabels,
)
from .retrieve_response_response_result_values import (
    RetrieveResponseResponseResultValues,
)
from .schema_property import SchemaProperty
from .schema_property_data_type import SchemaPropertyDataType
from .schema_property_format import SchemaPropertyFormat
from .schema_property_items import SchemaPropertyItems
from .schema_property_items_properties import SchemaPropertyItemsProperties
from .schema_property_items_properties_x import SchemaPropertyItemsPropertiesX
from .schema_property_items_properties_x_type import SchemaPropertyItemsPropertiesXType
from .schema_property_items_properties_y import SchemaPropertyItemsPropertiesY
from .schema_property_items_properties_y_type import SchemaPropertyItemsPropertiesYType
from .schema_property_type import SchemaPropertyType
from .survey_answer import SurveyAnswer
from .survey_not_found_response import SurveyNotFoundResponse
from .survey_unauthorized_response import SurveyUnauthorizedResponse
from .update_response_request import UpdateResponseRequest
from .update_response_request_embedded_data import UpdateResponseRequestEmbeddedData
from .update_response_response import UpdateResponseResponse

__all__ = (
    "CreateResponseRequest",
    "CreateResponseRequestValues",
    "CreateResponseResponse",
    "CreateResponseResponseResult",
    "CreateResponseWithFileAttachmentsRequest",
    "CreateResponseWithFileAttachmentsRequestFileMapping",
    "DefaultErrorResponse",
    "DeleteResponseResponse",
    "Error",
    "GetSchemaResponse",
    "GetSchemaResponseResult",
    "GetSchemaResponseResultProperties",
    "GetSchemaResponseResultPropertiesDisplayedFields",
    "GetSchemaResponseResultPropertiesDisplayedFieldsItems",
    "GetSchemaResponseResultPropertiesDisplayedFieldsItemsType",
    "GetSchemaResponseResultPropertiesDisplayedFieldsType",
    "GetSchemaResponseResultPropertiesDisplayedValues",
    "GetSchemaResponseResultPropertiesDisplayedValuesAdditionalProperties",
    "GetSchemaResponseResultPropertiesDisplayedValuesAdditionalPropertiesType",
    "GetSchemaResponseResultPropertiesDisplayedValuesType",
    "GetSchemaResponseResultPropertiesLabels",
    "GetSchemaResponseResultPropertiesLabelsAdditionalProperties",
    "GetSchemaResponseResultPropertiesLabelsType",
    "GetSchemaResponseResultPropertiesResponseId",
    "GetSchemaResponseResultPropertiesResponseIdType",
    "GetSchemaResponseResultPropertiesValues",
    "GetSchemaResponseResultPropertiesValuesProperties",
    "GetSchemaResponseResultPropertiesValuesType",
    "GetSchemaResponseResultSchema",
    "GetSchemaResponseResultType",
    "Meta",
    "MetaWithError",
    "RetrieveResponseResponse",
    "RetrieveResponseResponseNotYetAvailable",
    "RetrieveResponseResponseNotYetAvailableResult",
    "RetrieveResponseResponseResult",
    "RetrieveResponseResponseResultDisplayedValues",
    "RetrieveResponseResponseResultLabels",
    "RetrieveResponseResponseResultValues",
    "SchemaProperty",
    "SchemaPropertyDataType",
    "SchemaPropertyFormat",
    "SchemaPropertyItems",
    "SchemaPropertyItemsProperties",
    "SchemaPropertyItemsPropertiesX",
    "SchemaPropertyItemsPropertiesXType",
    "SchemaPropertyItemsPropertiesY",
    "SchemaPropertyItemsPropertiesYType",
    "SchemaPropertyType",
    "SurveyAnswer",
    "SurveyNotFoundResponse",
    "SurveyUnauthorizedResponse",
    "UpdateResponseRequest",
    "UpdateResponseRequestEmbeddedData",
    "UpdateResponseResponse",
)



import argparse
import json
import pathlib
from typing import *

from qualtrics_utils.codebook.generate import generate_codebook
from qualtrics_utils.utils import normalize_html_string, quote_value

GENERATOR_TYPES = ["tableau", "sql"]


def sql_qualtrics_map_func(mapping: dict) -> dict:
    question_number, question_string, answer_choices = (
        mapping["question_number"],
        mapping["question_string"],
        mapping.get("answer_choices", {}),
    )

    return {
        "key": question_number,
        "alias": question_string,
        "mapping": answer_choices,
    }


def generate_sql(table_name: str, key_mapping: dict) -> Tuple[str, str]:
    key = key_mapping.get("key", "")
    alias = key_mapping.get("alias", key)
    mapping = key_mapping.get("mapping", {})

    s = "CASE\n"

    for sub_key, sub_value in mapping.items():
        sub_key, sub_value = (
            quote_value(sub_key, '"'),
            quote_value(normalize_html_string(sub_value), '"'),
        )
        s += f"\tWHEN {key} = {sub_key} THEN {sub_value}\n"

    s += f"END AS {alias}"
    return key, s


def tableau_qualtrics_map_func(mapping: dict) -> dict:
    question_number, answer_choices = (
        mapping["question_number"],
        mapping.get("answer_choices", {}),
    )
    return {
        "key": question_number,
        "mapping": answer_choices,
    }


def generate_tableau(key_mapping: dict) -> Tuple[str, str]:
    key = key_mapping.get("key", "")
    mapping = key_mapping.get("mapping", {})

    s = ""
    for n, (sub_key, sub_value) in enumerate(mapping.items()):
        sub_key = sub_key if str(sub_key).isnumeric() else quote_value(sub_key, "'")
        sub_value = quote_value(normalize_html_string(sub_value), "'")

        s += "ELSE" if n > 0 else ""
        s += f"IF [{key}] == {sub_key} THEN "
        s += f"{sub_value} "

    s += f"END"
    return key, s


def get_map_and_generator_funcs(kind: str) -> tuple[Callable, Callable]:
    if kind == "sql":
        map_func = sql_qualtrics_map_func
        generator_func = lambda key_mapping: generate_sql("$$TABLE_NAME$$", key_mapping)
        return map_func, generator_func
    elif kind == "tableau":
        return tableau_qualtrics_map_func, generate_tableau

    raise ValueError(f"Invalid kind: {kind}")


def map_columns(
    input_filepath: pathlib.Path,
    kind: str,
) -> dict:
    map_func, generator_func = get_map_and_generator_funcs(kind)

    with open(input_filepath, "r") as file:
        mapping_list = (
            json.load(file)
            if input_filepath.suffix == ".json"
            else generate_codebook(input_filepath)
        )

        commands = {}
        for mapping in mapping_list:
            key_mapping = map_func(mapping)

            if key_mapping is None:
                continue

            key, command = generator_func(key_mapping)
            commands[key] = command

        return commands


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Input file path.", type=pathlib.Path)
    parser.add_argument(
        "--kind",
        help="Generated output kind.",
        choices=GENERATOR_TYPES,
        default=GENERATOR_TYPES[0],
    )

    args = parser.parse_args()

    filepath = args.input
    kind = args.kind
    out_path = filepath.parent / f"{filepath.name}-{kind}-map.json"

    commands = map_columns(filepath, kind)

    with open(out_path, "w") as file:
        json.dump(commands, file, indent=4)


if __name__ == "__main__":
    main()



from __future__ import annotations

import datetime
import urllib.parse
import zipfile
from io import BytesIO
from typing import IO, Any, Optional
from zipfile import ZipFile
import numpy as np

import pandas as pd
import requests
from loguru import logger

from qualtrics_utils.misc import BASE_URL, HEADERS, VERSION, ExportedFile
from qualtrics_utils.surveys_response_import_export_api_client.api.response_exports import (
    create_export,
    get_export_file,
    get_export_progress,
)
from qualtrics_utils.surveys_response_import_export_api_client.client import (
    AuthenticatedClient,
)
from qualtrics_utils.surveys_response_import_export_api_client.models import (
    ExportCreationRequest,
    ExportCreationRequestFormat,
    RequestStatus,
)
from qualtrics_utils.surveys_response_import_export_api_client.types import UNSET
from qualtrics_utils.utils import (
    parse_file_id,
    reset_request_defaults,
    qualtrics_schema_to_dtypes,
)

# The first two rows of the CSV are Qualtrics metadata.
SKIP_ROWS = [1, 2]


class Surveys:
    def __init__(self, api_token: str, version: str = VERSION):
        self.base_url = BASE_URL(version)

        self.headers = HEADERS.copy()
        self.headers["X-API-TOKEN"] = api_token

        self.version = version

        self.client = AuthenticatedClient(
            token=api_token,
            base_url=self.base_url,
            prefix="",
            auth_header_name="X-API-TOKEN",
        )

    @staticmethod
    def _get_zip(url: str) -> IO[bytes] | list[IO[bytes]]:
        r = requests.get(url)
        with BytesIO(r.content) as data, ZipFile(data) as zipfile:
            files = [zipfile.open(file_name) for file_name in zipfile.namelist()]
            return files.pop() if len(files) == 1 else files

    def _make_api_url(self, url: str, **kwargs: Any) -> str:
        return urllib.parse.urljoin(self.base_url, url.format(**kwargs))

    def _response_export(
        self,
        survey_id: str,
        format: ExportCreationRequestFormat = ExportCreationRequestFormat.CSV,
        use_labels: bool = True,
        start_date: datetime.datetime | None = None,
        end_date: datetime.datetime | None = None,
        export_responses_in_progress: bool = False,
        continuation_token: Optional[str] = None,
        **kwargs: Any,
    ):
        kwargs = dict(
            format_=format,
            use_labels=use_labels,
            breakout_sets=True,
            seen_unanswered_recode=-1,
            multiselect_seen_unanswered_recode=-1,
            allow_continuation=not export_responses_in_progress and end_date is None,
            sort_by_last_modified_date=False,
            include_label_columns=not use_labels,
            compress=True,
            export_responses_in_progress=export_responses_in_progress,
            start_date=start_date if start_date is not None else UNSET,
            end_date=end_date if end_date is not None else UNSET,
            continuation_token=continuation_token
            if continuation_token is not None
            else UNSET,
            **kwargs,
        )
        payload = ExportCreationRequest(**kwargs)
        # ! This is a hack: the Qualtrics OpenAPI docs suck tremendously and the defaults are NOT correct.
        payload = reset_request_defaults(payload, kwargs)

        logger.info(
            f"Exporting responses from {survey_id} from {start_date} to {end_date}."
        )
        logger.debug(f"Exporting with payload: {payload}")

        return create_export.sync(
            survey_id=survey_id,
            client=self.client,
            json_body=payload,
        )

    def _response_export_status(self, survey_id: str, export_progress_id: str):
        logger.info(f"Getting export progress for survey {survey_id}...")

        while True:
            r = get_export_progress.sync(
                survey_id=survey_id,
                export_progress_id=export_progress_id,
                client=self.client,
            )
            status = r.result.status  # type: ignore

            logger.info(
                f"Export progress for file {r.result.file_id}: {r.result.percent_complete}%"  # type: ignore
            )

            match status:
                case None | RequestStatus.FAILED:
                    raise Exception("Export failed", r)
                case RequestStatus.COMPLETE:
                    return r
                case RequestStatus.INPROGRESS:
                    continue

    def _response_export_file(self, survey_id: str, file_id: str) -> bytes:
        logger.info(f"Downloading file {file_id} for survey {survey_id}...")

        r = get_export_file.sync(
            survey_id=survey_id,
            file_id=file_id,
            client=self.client,
        )

        return r.payload.read()  # type: ignore

    def get_responses(
        self,
        survey_id: str,
        format: ExportCreationRequestFormat = ExportCreationRequestFormat.CSV,
        use_labels: bool = True,
        start_date: datetime.datetime | None = None,
        end_date: datetime.datetime | None = None,
        export_responses_in_progress: bool = False,
        continuation_token: Optional[str] = None,
        last_response_id: Optional[str] = None,
        **kwargs: Any,
    ) -> ExportedFile[bytes]:
        """Get responses from a survey by survey_id.
        Outputs a zipped file, in bytes, in the format specified by `format`.

        If a `last_response_id` is provided, the export will continue from the response after the last_response_id.
        If a `continuation_token` is provided, the export will continue from where it left off. The continuation_token **cannot** be older than 1 week.

        Args:
            survey_id (str): The survey_id of the survey to get responses from.
            format (str, optional): The format of the response data. Defaults to "csv".
            use_labels (bool, optional): Whether to use labels for the column names. Defaults to True.
            end_date (Optional[datetime.datetime], optional): The end date for the response export. Defaults to None.
            start_date (Optional[datetime.datetime], optional): The start date for the response export. Defaults to None.
            export_responses_in_progress (bool, optional): Whether to export responses that are in progress. Defaults to False.
            continuation_token (Optional[str], optional): The continuation token for the response export. Defaults to None.
            last_response_id (Optional[str], optional): The responseId of the last response to export. Defaults to None.
        """
        survey_id = parse_file_id(survey_id)

        start_date = (
            self._response_id_to_date(survey_id=survey_id, response_id=last_response_id)
            if start_date is None and last_response_id is not None
            else start_date
        )

        export = self._response_export(
            survey_id=survey_id,
            format=format,
            use_labels=use_labels,
            start_date=start_date,
            end_date=end_date,
            export_responses_in_progress=export_responses_in_progress,
            continuation_token=continuation_token,
            **kwargs,
        )

        progress_id = export.result.progress_id
        export_status = self._response_export_status(
            survey_id=survey_id, export_progress_id=progress_id
        )

        file_id = export_status.result.file_id
        file = self._response_export_file(survey_id=survey_id, file_id=file_id)

        return ExportedFile(
            survey_id=survey_id,
            file_id=file_id,
            last_response_id=None,
            continuation_token=export_status.result.continuation_token,
            data=file,
            timestamp=datetime.datetime.now(),
        )

    def get_response(self, survey_id: str, response_id: str) -> dict[str, Any]:
        """Get a single response from a survey by survey_id and response_id.

        Args:
            survey_id (str): The survey_id of the survey to get the response from.
            response_id (str): The response_id of the response to get.
        """
        survey_id = parse_file_id(survey_id)

        response_url = self._make_api_url(
            "surveys/{survey_id}/responses/{response_id}",
            survey_id=survey_id,
            response_id=response_id,
        )
        r = requests.get(response_url, headers=self.headers)
        r.raise_for_status()
        return r.json()

    def _response_id_to_date(
        self, survey_id: str, response_id: str
    ) -> datetime.datetime | None:
        response = self.get_response(survey_id=survey_id, response_id=response_id)

        return datetime.datetime.fromisoformat(
            response["result"]["values"]["startDate"]
        )

    def get_responses_df(
        self,
        survey_id: str,
        use_labels: bool = True,
        end_date: datetime.datetime | None = None,
        start_date: datetime.datetime | None = None,
        export_responses_in_progress: bool = False,
        continuation_token: Optional[str] = None,
        last_response_id: Optional[str] = None,
        filter_preview: bool = True,
        dtypes: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> ExportedFile[pd.DataFrame]:
        """Get responses from a survey by survey_id.
        Outputs a pandas DataFrame, with the index set to `ResponseId`.

        All blank values therein are replaced with pd.NA.,
        and all columns that are entirely pd.NA are cast to an object.

        If a last_response_id is provided, the export will continue from the response after the last_response_id.
        If a continuation_token is provided, the export will continue from where it left off. The continuation_token **cannot** be older than 1 week.

        Additional keyword arguments are passed to `get_responses`, see the ExportCreationRequest model for more details.

        Args:
            survey_id (str): The survey_id of the survey to get responses from.
            use_labels (bool, optional): Whether to use labels for the column names. Defaults to True.
            end_date (Optional[datetime.datetime], optional): The end date for the response export. Defaults to None.
            start_date (Optional[datetime.datetime], optional): The start date for the response export. Defaults to None.
            export_responses_in_progress (bool, optional): Whether to export responses that are in progress. Defaults to False.
            continuation_token (Optional[str], optional): The continuation token for the response export. Defaults to None.
            last_response_id (Optional[str], optional): The responseId of the last response to export. Defaults to None.
            filter_preview (bool, optional): Whether to filter out Survey Preview responses. Defaults to True.
            parse_dates (list[str], optional): List of columns to parse as dates. Defaults to ["StartDate", "EndDate"].
        """
        survey_id = parse_file_id(survey_id)

        raw_data = self.get_responses(
            survey_id=survey_id,
            use_labels=use_labels,
            end_date=end_date,
            start_date=start_date,
            export_responses_in_progress=export_responses_in_progress,
            continuation_token=continuation_token,
            last_response_id=last_response_id,
            **kwargs,
        )

        schema = self.get_survey_schema(survey_id=survey_id)
        dtypes = qualtrics_schema_to_dtypes(
            schema=schema, use_labels=use_labels, dtypes=dtypes
        )

        # Pandas' CSV reader cannot parse dates using the dtypes arg
        # So we have to feed them into the parse_dates arg
        parse_dates = []
        for col, type in dtypes.items():
            if isinstance(type, np.datetime64) or type == np.datetime64:
                parse_dates.append(col)

        dtypes = {k: v for k, v in dtypes.items() if k not in parse_dates}

        with zipfile.ZipFile(BytesIO(raw_data.data)) as data:
            with data.open(data.filelist[0]) as f:
                logger.info(f"Reading file {f.name}...")

                new_df_reader = pd.read_csv(
                    f,
                    skiprows=SKIP_ROWS,
                    iterator=True,
                    dtype=dtypes,
                    skip_blank_lines=True,
                    parse_dates=parse_dates,
                )
                new_df = pd.concat(new_df_reader, ignore_index=True)

                logger.info(f"Exported {len(new_df)} responses.")

                new_df.set_index("ResponseId", inplace=True)

                # If the last response is the same as the last response from the previous export, drop it.
                if not new_df.empty and new_df.index[0] == last_response_id:
                    new_df.drop(new_df.index[0], inplace=True)

                # Sort by StartDate
                new_df.sort_values("StartDate", inplace=True)

                # Replace all blank values with pd.NA
                try:
                    new_df.replace(
                        [r"^\s*$", "-1", -1], pd.NA, regex=True, inplace=True
                    )
                except Exception as e:
                    logger.error(e)

                # Cast all columns that are entirely pd.NA to object (str)
                new_df = new_df.astype(
                    {
                        col: "object"
                        for col in new_df.columns
                        if new_df[col].isna().all()
                    }
                )
                # Filter out Survey Preview responses
                if filter_preview and "Status" in new_df.columns:
                    preview_df = new_df[new_df["Status"] == "Survey Preview"]
                    logger.info(
                        f"Filtering out {len(preview_df)} Survey Preview responses."
                    )
                    new_df.drop(preview_df.index, inplace=True)

                # Set the last_response_id to the last response in the DataFrame, or the last_response_id from the previous export.
                last_response_id = (
                    new_df.index[-1] if len(new_df) > 0 else last_response_id
                )

                return ExportedFile(
                    survey_id=raw_data.survey_id,
                    file_id=raw_data.file_id,
                    last_response_id=last_response_id,
                    continuation_token=raw_data.continuation_token,
                    data=new_df,
                    timestamp=datetime.datetime.now(),
                )

    def get_survey_schema(self, survey_id: str) -> dict[str, Any]:
        """Get the schema of a survey by survey_id.

        Args:
            survey_id (str): The survey_id of the survey to get the schema from.
        """
        survey_id = parse_file_id(survey_id)

        logger.info(f"Getting schema for survey {survey_id}...")

        schema_url = self._make_api_url(
            "surveys/{survey_id}/response-schema/", survey_id=survey_id
        )
        r = requests.get(schema_url, headers=self.headers)
        r.raise_for_status()
        return r.json()



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.get_schema_response_result_properties_displayed_fields import (
        GetSchemaResponseResultPropertiesDisplayedFields,
    )
    from ..models.get_schema_response_result_properties_displayed_values import (
        GetSchemaResponseResultPropertiesDisplayedValues,
    )
    from ..models.get_schema_response_result_properties_labels import (
        GetSchemaResponseResultPropertiesLabels,
    )
    from ..models.get_schema_response_result_properties_response_id import (
        GetSchemaResponseResultPropertiesResponseId,
    )
    from ..models.get_schema_response_result_properties_values import (
        GetSchemaResponseResultPropertiesValues,
    )


T = TypeVar("T", bound="GetSchemaResponseResultProperties")


@_attrs_define
class GetSchemaResponseResultProperties:
    """
    Attributes:
        displayed_fields (Union[Unset, GetSchemaResponseResultPropertiesDisplayedFields]):
        displayed_values (Union[Unset, GetSchemaResponseResultPropertiesDisplayedValues]):
        labels (Union[Unset, GetSchemaResponseResultPropertiesLabels]):
        response_id (Union[Unset, GetSchemaResponseResultPropertiesResponseId]):
        values (Union[Unset, GetSchemaResponseResultPropertiesValues]):
    """

    displayed_fields: Union[
        Unset, "GetSchemaResponseResultPropertiesDisplayedFields"
    ] = UNSET
    displayed_values: Union[
        Unset, "GetSchemaResponseResultPropertiesDisplayedValues"
    ] = UNSET
    labels: Union[Unset, "GetSchemaResponseResultPropertiesLabels"] = UNSET
    response_id: Union[Unset, "GetSchemaResponseResultPropertiesResponseId"] = UNSET
    values: Union[Unset, "GetSchemaResponseResultPropertiesValues"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        displayed_fields: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.displayed_fields, Unset):
            displayed_fields = self.displayed_fields.to_dict()

        displayed_values: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.displayed_values, Unset):
            displayed_values = self.displayed_values.to_dict()

        labels: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.labels, Unset):
            labels = self.labels.to_dict()

        response_id: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.response_id, Unset):
            response_id = self.response_id.to_dict()

        values: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.values, Unset):
            values = self.values.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if displayed_fields is not UNSET:
            field_dict["displayedFields"] = displayed_fields
        if displayed_values is not UNSET:
            field_dict["displayedValues"] = displayed_values
        if labels is not UNSET:
            field_dict["labels"] = labels
        if response_id is not UNSET:
            field_dict["responseId"] = response_id
        if values is not UNSET:
            field_dict["values"] = values

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.get_schema_response_result_properties_displayed_fields import (
            GetSchemaResponseResultPropertiesDisplayedFields,
        )
        from ..models.get_schema_response_result_properties_displayed_values import (
            GetSchemaResponseResultPropertiesDisplayedValues,
        )
        from ..models.get_schema_response_result_properties_labels import (
            GetSchemaResponseResultPropertiesLabels,
        )
        from ..models.get_schema_response_result_properties_response_id import (
            GetSchemaResponseResultPropertiesResponseId,
        )
        from ..models.get_schema_response_result_properties_values import (
            GetSchemaResponseResultPropertiesValues,
        )

        d = src_dict.copy()
        _displayed_fields = d.pop("displayedFields", UNSET)
        displayed_fields: Union[Unset, GetSchemaResponseResultPropertiesDisplayedFields]
        if isinstance(_displayed_fields, Unset):
            displayed_fields = UNSET
        else:
            displayed_fields = (
                GetSchemaResponseResultPropertiesDisplayedFields.from_dict(
                    _displayed_fields
                )
            )

        _displayed_values = d.pop("displayedValues", UNSET)
        displayed_values: Union[Unset, GetSchemaResponseResultPropertiesDisplayedValues]
        if isinstance(_displayed_values, Unset):
            displayed_values = UNSET
        else:
            displayed_values = (
                GetSchemaResponseResultPropertiesDisplayedValues.from_dict(
                    _displayed_values
                )
            )

        _labels = d.pop("labels", UNSET)
        labels: Union[Unset, GetSchemaResponseResultPropertiesLabels]
        if isinstance(_labels, Unset):
            labels = UNSET
        else:
            labels = GetSchemaResponseResultPropertiesLabels.from_dict(_labels)

        _response_id = d.pop("responseId", UNSET)
        response_id: Union[Unset, GetSchemaResponseResultPropertiesResponseId]
        if isinstance(_response_id, Unset):
            response_id = UNSET
        else:
            response_id = GetSchemaResponseResultPropertiesResponseId.from_dict(
                _response_id
            )

        _values = d.pop("values", UNSET)
        values: Union[Unset, GetSchemaResponseResultPropertiesValues]
        if isinstance(_values, Unset):
            values = UNSET
        else:
            values = GetSchemaResponseResultPropertiesValues.from_dict(_values)

        get_schema_response_result_properties = cls(
            displayed_fields=displayed_fields,
            displayed_values=displayed_values,
            labels=labels,
            response_id=response_id,
            values=values,
        )

        get_schema_response_result_properties.additional_properties = d
        return get_schema_response_result_properties

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.error import Error


T = TypeVar("T", bound="MetaWithError")


@_attrs_define
class MetaWithError:
    """
    Attributes:
        http_status (str): HTTP status code
        request_id (str): An identifier for the incoming request.
        error (Error):
        notice (Union[Unset, str]):
    """

    http_status: str
    request_id: str
    error: "Error"
    notice: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        http_status = self.http_status
        request_id = self.request_id
        error = self.error.to_dict()

        notice = self.notice

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "httpStatus": http_status,
                "requestId": request_id,
                "error": error,
            }
        )
        if notice is not UNSET:
            field_dict["notice"] = notice

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.error import Error

        d = src_dict.copy()
        http_status = d.pop("httpStatus")

        request_id = d.pop("requestId")

        error = Error.from_dict(d.pop("error"))

        notice = d.pop("notice", UNSET)

        meta_with_error = cls(
            http_status=http_status,
            request_id=request_id,
            error=error,
            notice=notice,
        )

        meta_with_error.additional_properties = d
        return meta_with_error

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.create_import_json_body import CreateImportJsonBody
from ...models.creation_response import CreationResponse
from ...models.creation_response_invalid import CreationResponseInvalid
from ...models.creation_response_too_large import CreationResponseTooLarge
from ...models.internal_error import InternalError
from ...models.meta import Meta
from ...models.survey_not_found import SurveyNotFound
from ...types import Response, Unset


def _get_kwargs(
    survey_id: str,
    *,
    json_body: CreateImportJsonBody,
    idempotency_key: Union[Unset, str] = "",
) -> Dict[str, Any]:
    headers = {}
    if not isinstance(idempotency_key, Unset):
        headers["Idempotency-Key"] = idempotency_key

    json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": "/surveys/{surveyId}/import-responses".format(
            surveyId=survey_id,
        ),
        "json": json_json_body,
        "headers": headers,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[
        CreationResponse,
        CreationResponseInvalid,
        CreationResponseTooLarge,
        InternalError,
        Meta,
        SurveyNotFound,
    ]
]:
    if response.status_code == HTTPStatus.OK:
        response_200 = CreationResponse.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = CreationResponseInvalid.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = Meta.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = SurveyNotFound.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.REQUEST_ENTITY_TOO_LARGE:
        response_413 = CreationResponseTooLarge.from_dict(response.json())

        return response_413
    if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
        response_429 = CreationResponseTooLarge.from_dict(response.json())

        return response_429
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = InternalError.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[
        CreationResponse,
        CreationResponseInvalid,
        CreationResponseTooLarge,
        InternalError,
        Meta,
        SurveyNotFound,
    ]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    json_body: CreateImportJsonBody,
    idempotency_key: Union[Unset, str] = "",
) -> Response[
    Union[
        CreationResponse,
        CreationResponseInvalid,
        CreationResponseTooLarge,
        InternalError,
        Meta,
        SurveyNotFound,
    ]
]:
    r""" Start Response Import

     Starts an import of a CSV or TSV file. See the [Response Import/Export API
    Overview](../../../../reference/responseImportsExports.json) for more detail on how to use this
    endpoint within an import workflow.

    <!-- theme: warning -->
    > ### The API console does not generate code correctly for this endpoint. Use curl for sending
    requests.
    > The API console does not render code generation correctly when used to upload CSV or TSV files to
    this endpoint. Use curl for testing this endpoint.


    ### Import via Locally Hosted File
    To import using a locally hosted CSV or TSV file, see this example:

    <!--
    title: Import via Locally Hosted File
    -->
    ```shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
      --data-binary '@MyResponses.csv'--header 'Content-Type: text/csv' \
      --header 'charset: UTF-8' \
      --header 'Authorization: Bearer ec9f2045-e33d-4201-80c7-beee28728ef6'
    ```

    ### Import via a File Hosted at a Public URL
    To import using a remote, publicly hosted CSV or TSV file, see this example:

    <!--
    title: Import via a File Hosted at a Public URL
    lineNumbers: true
    -->
    ``` shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
        --header 'Content-Type: application/json' \
        --header 'Authorization: Bearer 8c1e372c-b65a-4ef1-b761-51affbc80216' \
        --data-raw '{
            \"format\": \"csv\",
            \"fileUrl\": \"http://myfilehost.com/MyResponses.csv\"
        }'
    ```

    <!-- theme: danger -->
    > ### Files Uploaded from a Remote Location Require a Public URL
    > A CSV or TSV file uploaded from a remote location must be available via a publicly accessible URL.

    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):  Default: ''.
        json_body (CreateImportJsonBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[CreationResponse, CreationResponseInvalid, CreationResponseTooLarge, InternalError, Meta, SurveyNotFound]]
     """

    kwargs = _get_kwargs(
        survey_id=survey_id,
        json_body=json_body,
        idempotency_key=idempotency_key,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    json_body: CreateImportJsonBody,
    idempotency_key: Union[Unset, str] = "",
) -> Optional[
    Union[
        CreationResponse,
        CreationResponseInvalid,
        CreationResponseTooLarge,
        InternalError,
        Meta,
        SurveyNotFound,
    ]
]:
    r""" Start Response Import

     Starts an import of a CSV or TSV file. See the [Response Import/Export API
    Overview](../../../../reference/responseImportsExports.json) for more detail on how to use this
    endpoint within an import workflow.

    <!-- theme: warning -->
    > ### The API console does not generate code correctly for this endpoint. Use curl for sending
    requests.
    > The API console does not render code generation correctly when used to upload CSV or TSV files to
    this endpoint. Use curl for testing this endpoint.


    ### Import via Locally Hosted File
    To import using a locally hosted CSV or TSV file, see this example:

    <!--
    title: Import via Locally Hosted File
    -->
    ```shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
      --data-binary '@MyResponses.csv'--header 'Content-Type: text/csv' \
      --header 'charset: UTF-8' \
      --header 'Authorization: Bearer ec9f2045-e33d-4201-80c7-beee28728ef6'
    ```

    ### Import via a File Hosted at a Public URL
    To import using a remote, publicly hosted CSV or TSV file, see this example:

    <!--
    title: Import via a File Hosted at a Public URL
    lineNumbers: true
    -->
    ``` shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
        --header 'Content-Type: application/json' \
        --header 'Authorization: Bearer 8c1e372c-b65a-4ef1-b761-51affbc80216' \
        --data-raw '{
            \"format\": \"csv\",
            \"fileUrl\": \"http://myfilehost.com/MyResponses.csv\"
        }'
    ```

    <!-- theme: danger -->
    > ### Files Uploaded from a Remote Location Require a Public URL
    > A CSV or TSV file uploaded from a remote location must be available via a publicly accessible URL.

    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):  Default: ''.
        json_body (CreateImportJsonBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[CreationResponse, CreationResponseInvalid, CreationResponseTooLarge, InternalError, Meta, SurveyNotFound]
     """

    return sync_detailed(
        survey_id=survey_id,
        client=client,
        json_body=json_body,
        idempotency_key=idempotency_key,
    ).parsed


async def asyncio_detailed(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    json_body: CreateImportJsonBody,
    idempotency_key: Union[Unset, str] = "",
) -> Response[
    Union[
        CreationResponse,
        CreationResponseInvalid,
        CreationResponseTooLarge,
        InternalError,
        Meta,
        SurveyNotFound,
    ]
]:
    r""" Start Response Import

     Starts an import of a CSV or TSV file. See the [Response Import/Export API
    Overview](../../../../reference/responseImportsExports.json) for more detail on how to use this
    endpoint within an import workflow.

    <!-- theme: warning -->
    > ### The API console does not generate code correctly for this endpoint. Use curl for sending
    requests.
    > The API console does not render code generation correctly when used to upload CSV or TSV files to
    this endpoint. Use curl for testing this endpoint.


    ### Import via Locally Hosted File
    To import using a locally hosted CSV or TSV file, see this example:

    <!--
    title: Import via Locally Hosted File
    -->
    ```shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
      --data-binary '@MyResponses.csv'--header 'Content-Type: text/csv' \
      --header 'charset: UTF-8' \
      --header 'Authorization: Bearer ec9f2045-e33d-4201-80c7-beee28728ef6'
    ```

    ### Import via a File Hosted at a Public URL
    To import using a remote, publicly hosted CSV or TSV file, see this example:

    <!--
    title: Import via a File Hosted at a Public URL
    lineNumbers: true
    -->
    ``` shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
        --header 'Content-Type: application/json' \
        --header 'Authorization: Bearer 8c1e372c-b65a-4ef1-b761-51affbc80216' \
        --data-raw '{
            \"format\": \"csv\",
            \"fileUrl\": \"http://myfilehost.com/MyResponses.csv\"
        }'
    ```

    <!-- theme: danger -->
    > ### Files Uploaded from a Remote Location Require a Public URL
    > A CSV or TSV file uploaded from a remote location must be available via a publicly accessible URL.

    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):  Default: ''.
        json_body (CreateImportJsonBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[CreationResponse, CreationResponseInvalid, CreationResponseTooLarge, InternalError, Meta, SurveyNotFound]]
     """

    kwargs = _get_kwargs(
        survey_id=survey_id,
        json_body=json_body,
        idempotency_key=idempotency_key,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    json_body: CreateImportJsonBody,
    idempotency_key: Union[Unset, str] = "",
) -> Optional[
    Union[
        CreationResponse,
        CreationResponseInvalid,
        CreationResponseTooLarge,
        InternalError,
        Meta,
        SurveyNotFound,
    ]
]:
    r""" Start Response Import

     Starts an import of a CSV or TSV file. See the [Response Import/Export API
    Overview](../../../../reference/responseImportsExports.json) for more detail on how to use this
    endpoint within an import workflow.

    <!-- theme: warning -->
    > ### The API console does not generate code correctly for this endpoint. Use curl for sending
    requests.
    > The API console does not render code generation correctly when used to upload CSV or TSV files to
    this endpoint. Use curl for testing this endpoint.


    ### Import via Locally Hosted File
    To import using a locally hosted CSV or TSV file, see this example:

    <!--
    title: Import via Locally Hosted File
    -->
    ```shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
      --data-binary '@MyResponses.csv'--header 'Content-Type: text/csv' \
      --header 'charset: UTF-8' \
      --header 'Authorization: Bearer ec9f2045-e33d-4201-80c7-beee28728ef6'
    ```

    ### Import via a File Hosted at a Public URL
    To import using a remote, publicly hosted CSV or TSV file, see this example:

    <!--
    title: Import via a File Hosted at a Public URL
    lineNumbers: true
    -->
    ``` shell
    curl --location --request POST 'iad1.qualtrics.com/API/v3/surveys/SV_bwrylOA5nNnI9M1/import-
    responses' \
        --header 'Content-Type: application/json' \
        --header 'Authorization: Bearer 8c1e372c-b65a-4ef1-b761-51affbc80216' \
        --data-raw '{
            \"format\": \"csv\",
            \"fileUrl\": \"http://myfilehost.com/MyResponses.csv\"
        }'
    ```

    <!-- theme: danger -->
    > ### Files Uploaded from a Remote Location Require a Public URL
    > A CSV or TSV file uploaded from a remote location must be available via a publicly accessible URL.

    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):  Default: ''.
        json_body (CreateImportJsonBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[CreationResponse, CreationResponseInvalid, CreationResponseTooLarge, InternalError, Meta, SurveyNotFound]
     """

    return (
        await asyncio_detailed(
            survey_id=survey_id,
            client=client,
            json_body=json_body,
            idempotency_key=idempotency_key,
        )
    ).parsed



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.default_error_response import DefaultErrorResponse
from ...models.get_schema_response import GetSchemaResponse
from ...models.survey_not_found_response import SurveyNotFoundResponse
from ...types import Response


def _get_kwargs(
    survey_id: str,
) -> Dict[str, Any]:

    pass

    return {
        "method": "get",
        "url": "/surveys/{surveyId}/response-schema".format(
            surveyId=survey_id,
        ),
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = GetSchemaResponse.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = SurveyNotFoundResponse.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = DefaultErrorResponse.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    survey_id: str,
    *,
    client: AuthenticatedClient,
) -> Response[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]:
    r"""Retrieve Survey JSON Schema

     ## Introduction

    This documentation is meant to help explain the purpose and value of the Retrieve the JSON Schema
    endpoint in the public api. In short, this endpoint provides a [standard JSON Schema](https://json-
    schema.org/) that can be used for parsing the results from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint
    and the [Create Response Export](./legacyResponseImportsExports.json/paths/~1responseexports/post)
    endpoint when exporting in the json format. Finally this endpoint can also be useful for preparing
    responses for the [Create a New
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses/post) endpoint.

    # Example

    Note, before reading through this example, it might be helpful to take a quick detour and gain a bit
    of familiarity around Json Schemas.
    For this example we created a simple survey with a single multiple choice question `Which type of
    Ice Cream is your favorite?` and 4 options `Chocolate`, `Vanilla`, `Cookies & Cream`, and
    `Strawberry`.
    Here's an example response that we requested from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint:

    ```json
    {
        \"responseId\": \"R_240YEnuWbbwZOy1\",
        \"values\": {
            \"startDate\": \"2020-05-13T19:26:15Z\",
            \"endDate\": \"2020-05-13T19:26:18Z\",
            \"status\": 0,
            \"ipAddress\": \"70.103.180.85\",
            \"progress\": 100,
            \"duration\": 2,
            \"finished\": 1,
            \"recordedDate\": \"2020-05-13T19:26:18.487Z\",
            \"locationLatitude\": \"40.3941955566\",
            \"locationLongitude\": \"-111.848297119\",
            \"distributionChannel\": \"anonymous\",
            \"userLanguage\": \"EN\",
            \"QID1\": 3,
            \"QID1_DO\": [
                \"1\",
                \"2\",
                \"3\",
                \"4\"
            ]
        },
        \"labels\": {
            \"status\": \"IP Address\",
            \"finished\": \"True\",
            \"QID1\": \"Cookies & Cream\",
            \"QID1_DO\": [
                \"Chocolate\",
                \"Vanilla\",
                \"Cookies & Cream\",
                \"Strawberry\"
            ]
        },
        \"displayedFields\": [
            \"QID1\"
        ],
        \"displayedValues\": {
            \"QID1\": [
                1,
                2,
                3,
                4
            ]
        }
    }
    ```

    Now, let's look at the pieces of the schema to understand them a little better. Note: this
    documentation won't attempt to cover all the aspects of the JSON Schema specification, for more
    information see [https://json-schema.org/](https://json-schema.org/).

    ```json
    {
        \"$schema\": \"https://json-schema.org/draft/2019-09/schema\",
        \"$id\": \"http://qualtrics.com/API/v3/surveys/SV_ai71yqu2ZLvpadf/response-schema\",
        \"title\": \"JSON Schema Example (SV_ai71yqu2ZLvpadf) - Survey Response\",
        \"description\": \"This is a schema for survey responses from SV_ai71yqu2ZLvpadf\",
        \"type\": \"object\",
    ```

    The schema starts with several standard fields from the Json Schema specification. The value for
    `$schema` is explained [here](https://json-schema.org/draft/2019-09/json-schema-
    core.html#rfc.section.8.1.1) and the value for `$id` is explained [here](https://json-
    schema.org/draft/2019-09/json-schema-core.html#id-keyword). Then, the schema is given a title and a
    description. Finally, the schema is defined for a JSON [object](https://json-
    schema.org/draft/2019-09/json-schema-core.html#rfc.section.4.2.1) primitive type.

    ```json
        \"required\": [
            \"values\",
            \"displayedFields\",
            \"displayedValues\",
            \"labels\",
            \"responseId\"
        ],
    ```

    Next, the `required` field specifies that our JSON `object` will have the certain required fields.
    Looking back at our example response, we see that each of these fields is present.

    ```json
        \"properties\": {
            \"responseId\": {
                \"type\": \"string\"
            },
    ```

    Inside of the `properties` object, each field previously mentioned in the `required` array are given
    a schema. The simplest of these fields is the responseId field which will just be provided as a
    string.

    ```json
            \"displayedFields\": {
                \"type\": \"array\",
                \"description\": \"A list of questions or question rows that were displayed to the
    survey respondent\",
                \"items\": {
                    \"type\": \"string\"
                }
            },
            \"displayedValues\": {
                \"type\": \"object\",
                \"description\": \"A list of the possible answers shown to the survey respondent for
    each question or question row\",
                \"additionalProperties\": {
                    \"type\": \"array\"
                }
            },
            \"labels\": {
                \"type\": \"object\",
                \"description\": \"The labels of the answers given by the survey respondent for each
    question or question row\",
                \"additionalProperties\": true
            },
    ```

    The next fields `displayedFields`, `displayedValues`, and `labels` have descriptions built in to the
    schema. Again, we can see how these schemas translate into a response object by consulting our
    example.

    ```json
            \"values\": {
                \"type\": \"object\",
                \"description\": \"The answers given by the survey respondent\",
                \"properties\": {
                    \"QID1\": {
                        \"description\": \"Which type of Ice Cream is your favorite?\",
                        \"dataType\": \"question\",
                        \"type\": \"number\",
                        \"oneOf\": [
                            {
                                \"label\": \"Chocolate\",
                                \"const\": 1
                            },
                            {
                                \"label\": \"Vanilla\",
                                \"const\": 2
                            },
                            {
                                \"label\": \"Cookies & Cream\",
                                \"const\": 3
                            },
                            {
                                \"label\": \"Strawberry\",
                                \"const\": 4
                            }
                        ],
                        \"exportTag\": \"Q1\",
                        \"questionId\": \"QID1\"
                    },
                    \"recordedDate\": {
                        \"description\": \"Recorded Date\",
                        \"dataType\": \"metadata\",
                        \"type\": \"string\",
                        \"format\": \"date-time\",
                        \"exportTag\": \"RecordedDate\"
                    },
                    \"duration\": {
                        \"description\": \"Duration (in seconds)\",
                        \"dataType\": \"metadata\",
                        \"type\": \"number\",
                        \"exportTag\": \"Duration (in seconds)\"
                    },
    ```

    Finally, the values property defines the schema for each value available in the response. I've
    pulled a couple of the fields from the schema into this documentation but we can see from the
    example response that there are many more fields in this schema. Here we can see a `description` for
    each field as well as a `type` and several other pieces of information. We can see that the `type`
    matches up with the JSON data type for each value as found in the example response. The `dataType`
    can be used to know the source for that value and the `exportTag` is useful for understanding the
    header columns when using the [Create Response
    Export](./responseImportsExports.json/paths/~1surveys~1{surveyId}~1export-responses/post) endpoint
    with the CSV or TSV formats. Lastly, oneOf enumerates the possible options for values that have a
    known, finite domain. For this example, a multiple choice question would have a oneOf field since
    there are only 4 possible values. However, a text entry field that only allows numbers would not
    have a oneOf field, since the domain can't be enumerated.

    ## Additional Tooling for JSON Schemas

    One of the benefits of having a schema is that it can be used to automate some of the parsing and
    usage of json schemas. One such tool for automating this step is
    [Quicktype](https://app.quicktype.io/#l=schema). More information about that tool can be found on
    [Github](https://github.com/quicktype/quicktype).


    Args:
        survey_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]
    """

    kwargs = _get_kwargs(
        survey_id=survey_id,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    survey_id: str,
    *,
    client: AuthenticatedClient,
) -> Optional[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]:
    r"""Retrieve Survey JSON Schema

     ## Introduction

    This documentation is meant to help explain the purpose and value of the Retrieve the JSON Schema
    endpoint in the public api. In short, this endpoint provides a [standard JSON Schema](https://json-
    schema.org/) that can be used for parsing the results from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint
    and the [Create Response Export](./legacyResponseImportsExports.json/paths/~1responseexports/post)
    endpoint when exporting in the json format. Finally this endpoint can also be useful for preparing
    responses for the [Create a New
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses/post) endpoint.

    # Example

    Note, before reading through this example, it might be helpful to take a quick detour and gain a bit
    of familiarity around Json Schemas.
    For this example we created a simple survey with a single multiple choice question `Which type of
    Ice Cream is your favorite?` and 4 options `Chocolate`, `Vanilla`, `Cookies & Cream`, and
    `Strawberry`.
    Here's an example response that we requested from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint:

    ```json
    {
        \"responseId\": \"R_240YEnuWbbwZOy1\",
        \"values\": {
            \"startDate\": \"2020-05-13T19:26:15Z\",
            \"endDate\": \"2020-05-13T19:26:18Z\",
            \"status\": 0,
            \"ipAddress\": \"70.103.180.85\",
            \"progress\": 100,
            \"duration\": 2,
            \"finished\": 1,
            \"recordedDate\": \"2020-05-13T19:26:18.487Z\",
            \"locationLatitude\": \"40.3941955566\",
            \"locationLongitude\": \"-111.848297119\",
            \"distributionChannel\": \"anonymous\",
            \"userLanguage\": \"EN\",
            \"QID1\": 3,
            \"QID1_DO\": [
                \"1\",
                \"2\",
                \"3\",
                \"4\"
            ]
        },
        \"labels\": {
            \"status\": \"IP Address\",
            \"finished\": \"True\",
            \"QID1\": \"Cookies & Cream\",
            \"QID1_DO\": [
                \"Chocolate\",
                \"Vanilla\",
                \"Cookies & Cream\",
                \"Strawberry\"
            ]
        },
        \"displayedFields\": [
            \"QID1\"
        ],
        \"displayedValues\": {
            \"QID1\": [
                1,
                2,
                3,
                4
            ]
        }
    }
    ```

    Now, let's look at the pieces of the schema to understand them a little better. Note: this
    documentation won't attempt to cover all the aspects of the JSON Schema specification, for more
    information see [https://json-schema.org/](https://json-schema.org/).

    ```json
    {
        \"$schema\": \"https://json-schema.org/draft/2019-09/schema\",
        \"$id\": \"http://qualtrics.com/API/v3/surveys/SV_ai71yqu2ZLvpadf/response-schema\",
        \"title\": \"JSON Schema Example (SV_ai71yqu2ZLvpadf) - Survey Response\",
        \"description\": \"This is a schema for survey responses from SV_ai71yqu2ZLvpadf\",
        \"type\": \"object\",
    ```

    The schema starts with several standard fields from the Json Schema specification. The value for
    `$schema` is explained [here](https://json-schema.org/draft/2019-09/json-schema-
    core.html#rfc.section.8.1.1) and the value for `$id` is explained [here](https://json-
    schema.org/draft/2019-09/json-schema-core.html#id-keyword). Then, the schema is given a title and a
    description. Finally, the schema is defined for a JSON [object](https://json-
    schema.org/draft/2019-09/json-schema-core.html#rfc.section.4.2.1) primitive type.

    ```json
        \"required\": [
            \"values\",
            \"displayedFields\",
            \"displayedValues\",
            \"labels\",
            \"responseId\"
        ],
    ```

    Next, the `required` field specifies that our JSON `object` will have the certain required fields.
    Looking back at our example response, we see that each of these fields is present.

    ```json
        \"properties\": {
            \"responseId\": {
                \"type\": \"string\"
            },
    ```

    Inside of the `properties` object, each field previously mentioned in the `required` array are given
    a schema. The simplest of these fields is the responseId field which will just be provided as a
    string.

    ```json
            \"displayedFields\": {
                \"type\": \"array\",
                \"description\": \"A list of questions or question rows that were displayed to the
    survey respondent\",
                \"items\": {
                    \"type\": \"string\"
                }
            },
            \"displayedValues\": {
                \"type\": \"object\",
                \"description\": \"A list of the possible answers shown to the survey respondent for
    each question or question row\",
                \"additionalProperties\": {
                    \"type\": \"array\"
                }
            },
            \"labels\": {
                \"type\": \"object\",
                \"description\": \"The labels of the answers given by the survey respondent for each
    question or question row\",
                \"additionalProperties\": true
            },
    ```

    The next fields `displayedFields`, `displayedValues`, and `labels` have descriptions built in to the
    schema. Again, we can see how these schemas translate into a response object by consulting our
    example.

    ```json
            \"values\": {
                \"type\": \"object\",
                \"description\": \"The answers given by the survey respondent\",
                \"properties\": {
                    \"QID1\": {
                        \"description\": \"Which type of Ice Cream is your favorite?\",
                        \"dataType\": \"question\",
                        \"type\": \"number\",
                        \"oneOf\": [
                            {
                                \"label\": \"Chocolate\",
                                \"const\": 1
                            },
                            {
                                \"label\": \"Vanilla\",
                                \"const\": 2
                            },
                            {
                                \"label\": \"Cookies & Cream\",
                                \"const\": 3
                            },
                            {
                                \"label\": \"Strawberry\",
                                \"const\": 4
                            }
                        ],
                        \"exportTag\": \"Q1\",
                        \"questionId\": \"QID1\"
                    },
                    \"recordedDate\": {
                        \"description\": \"Recorded Date\",
                        \"dataType\": \"metadata\",
                        \"type\": \"string\",
                        \"format\": \"date-time\",
                        \"exportTag\": \"RecordedDate\"
                    },
                    \"duration\": {
                        \"description\": \"Duration (in seconds)\",
                        \"dataType\": \"metadata\",
                        \"type\": \"number\",
                        \"exportTag\": \"Duration (in seconds)\"
                    },
    ```

    Finally, the values property defines the schema for each value available in the response. I've
    pulled a couple of the fields from the schema into this documentation but we can see from the
    example response that there are many more fields in this schema. Here we can see a `description` for
    each field as well as a `type` and several other pieces of information. We can see that the `type`
    matches up with the JSON data type for each value as found in the example response. The `dataType`
    can be used to know the source for that value and the `exportTag` is useful for understanding the
    header columns when using the [Create Response
    Export](./responseImportsExports.json/paths/~1surveys~1{surveyId}~1export-responses/post) endpoint
    with the CSV or TSV formats. Lastly, oneOf enumerates the possible options for values that have a
    known, finite domain. For this example, a multiple choice question would have a oneOf field since
    there are only 4 possible values. However, a text entry field that only allows numbers would not
    have a oneOf field, since the domain can't be enumerated.

    ## Additional Tooling for JSON Schemas

    One of the benefits of having a schema is that it can be used to automate some of the parsing and
    usage of json schemas. One such tool for automating this step is
    [Quicktype](https://app.quicktype.io/#l=schema). More information about that tool can be found on
    [Github](https://github.com/quicktype/quicktype).


    Args:
        survey_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]
    """

    return sync_detailed(
        survey_id=survey_id,
        client=client,
    ).parsed


async def asyncio_detailed(
    survey_id: str,
    *,
    client: AuthenticatedClient,
) -> Response[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]:
    r"""Retrieve Survey JSON Schema

     ## Introduction

    This documentation is meant to help explain the purpose and value of the Retrieve the JSON Schema
    endpoint in the public api. In short, this endpoint provides a [standard JSON Schema](https://json-
    schema.org/) that can be used for parsing the results from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint
    and the [Create Response Export](./legacyResponseImportsExports.json/paths/~1responseexports/post)
    endpoint when exporting in the json format. Finally this endpoint can also be useful for preparing
    responses for the [Create a New
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses/post) endpoint.

    # Example

    Note, before reading through this example, it might be helpful to take a quick detour and gain a bit
    of familiarity around Json Schemas.
    For this example we created a simple survey with a single multiple choice question `Which type of
    Ice Cream is your favorite?` and 4 options `Chocolate`, `Vanilla`, `Cookies & Cream`, and
    `Strawberry`.
    Here's an example response that we requested from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint:

    ```json
    {
        \"responseId\": \"R_240YEnuWbbwZOy1\",
        \"values\": {
            \"startDate\": \"2020-05-13T19:26:15Z\",
            \"endDate\": \"2020-05-13T19:26:18Z\",
            \"status\": 0,
            \"ipAddress\": \"70.103.180.85\",
            \"progress\": 100,
            \"duration\": 2,
            \"finished\": 1,
            \"recordedDate\": \"2020-05-13T19:26:18.487Z\",
            \"locationLatitude\": \"40.3941955566\",
            \"locationLongitude\": \"-111.848297119\",
            \"distributionChannel\": \"anonymous\",
            \"userLanguage\": \"EN\",
            \"QID1\": 3,
            \"QID1_DO\": [
                \"1\",
                \"2\",
                \"3\",
                \"4\"
            ]
        },
        \"labels\": {
            \"status\": \"IP Address\",
            \"finished\": \"True\",
            \"QID1\": \"Cookies & Cream\",
            \"QID1_DO\": [
                \"Chocolate\",
                \"Vanilla\",
                \"Cookies & Cream\",
                \"Strawberry\"
            ]
        },
        \"displayedFields\": [
            \"QID1\"
        ],
        \"displayedValues\": {
            \"QID1\": [
                1,
                2,
                3,
                4
            ]
        }
    }
    ```

    Now, let's look at the pieces of the schema to understand them a little better. Note: this
    documentation won't attempt to cover all the aspects of the JSON Schema specification, for more
    information see [https://json-schema.org/](https://json-schema.org/).

    ```json
    {
        \"$schema\": \"https://json-schema.org/draft/2019-09/schema\",
        \"$id\": \"http://qualtrics.com/API/v3/surveys/SV_ai71yqu2ZLvpadf/response-schema\",
        \"title\": \"JSON Schema Example (SV_ai71yqu2ZLvpadf) - Survey Response\",
        \"description\": \"This is a schema for survey responses from SV_ai71yqu2ZLvpadf\",
        \"type\": \"object\",
    ```

    The schema starts with several standard fields from the Json Schema specification. The value for
    `$schema` is explained [here](https://json-schema.org/draft/2019-09/json-schema-
    core.html#rfc.section.8.1.1) and the value for `$id` is explained [here](https://json-
    schema.org/draft/2019-09/json-schema-core.html#id-keyword). Then, the schema is given a title and a
    description. Finally, the schema is defined for a JSON [object](https://json-
    schema.org/draft/2019-09/json-schema-core.html#rfc.section.4.2.1) primitive type.

    ```json
        \"required\": [
            \"values\",
            \"displayedFields\",
            \"displayedValues\",
            \"labels\",
            \"responseId\"
        ],
    ```

    Next, the `required` field specifies that our JSON `object` will have the certain required fields.
    Looking back at our example response, we see that each of these fields is present.

    ```json
        \"properties\": {
            \"responseId\": {
                \"type\": \"string\"
            },
    ```

    Inside of the `properties` object, each field previously mentioned in the `required` array are given
    a schema. The simplest of these fields is the responseId field which will just be provided as a
    string.

    ```json
            \"displayedFields\": {
                \"type\": \"array\",
                \"description\": \"A list of questions or question rows that were displayed to the
    survey respondent\",
                \"items\": {
                    \"type\": \"string\"
                }
            },
            \"displayedValues\": {
                \"type\": \"object\",
                \"description\": \"A list of the possible answers shown to the survey respondent for
    each question or question row\",
                \"additionalProperties\": {
                    \"type\": \"array\"
                }
            },
            \"labels\": {
                \"type\": \"object\",
                \"description\": \"The labels of the answers given by the survey respondent for each
    question or question row\",
                \"additionalProperties\": true
            },
    ```

    The next fields `displayedFields`, `displayedValues`, and `labels` have descriptions built in to the
    schema. Again, we can see how these schemas translate into a response object by consulting our
    example.

    ```json
            \"values\": {
                \"type\": \"object\",
                \"description\": \"The answers given by the survey respondent\",
                \"properties\": {
                    \"QID1\": {
                        \"description\": \"Which type of Ice Cream is your favorite?\",
                        \"dataType\": \"question\",
                        \"type\": \"number\",
                        \"oneOf\": [
                            {
                                \"label\": \"Chocolate\",
                                \"const\": 1
                            },
                            {
                                \"label\": \"Vanilla\",
                                \"const\": 2
                            },
                            {
                                \"label\": \"Cookies & Cream\",
                                \"const\": 3
                            },
                            {
                                \"label\": \"Strawberry\",
                                \"const\": 4
                            }
                        ],
                        \"exportTag\": \"Q1\",
                        \"questionId\": \"QID1\"
                    },
                    \"recordedDate\": {
                        \"description\": \"Recorded Date\",
                        \"dataType\": \"metadata\",
                        \"type\": \"string\",
                        \"format\": \"date-time\",
                        \"exportTag\": \"RecordedDate\"
                    },
                    \"duration\": {
                        \"description\": \"Duration (in seconds)\",
                        \"dataType\": \"metadata\",
                        \"type\": \"number\",
                        \"exportTag\": \"Duration (in seconds)\"
                    },
    ```

    Finally, the values property defines the schema for each value available in the response. I've
    pulled a couple of the fields from the schema into this documentation but we can see from the
    example response that there are many more fields in this schema. Here we can see a `description` for
    each field as well as a `type` and several other pieces of information. We can see that the `type`
    matches up with the JSON data type for each value as found in the example response. The `dataType`
    can be used to know the source for that value and the `exportTag` is useful for understanding the
    header columns when using the [Create Response
    Export](./responseImportsExports.json/paths/~1surveys~1{surveyId}~1export-responses/post) endpoint
    with the CSV or TSV formats. Lastly, oneOf enumerates the possible options for values that have a
    known, finite domain. For this example, a multiple choice question would have a oneOf field since
    there are only 4 possible values. However, a text entry field that only allows numbers would not
    have a oneOf field, since the domain can't be enumerated.

    ## Additional Tooling for JSON Schemas

    One of the benefits of having a schema is that it can be used to automate some of the parsing and
    usage of json schemas. One such tool for automating this step is
    [Quicktype](https://app.quicktype.io/#l=schema). More information about that tool can be found on
    [Github](https://github.com/quicktype/quicktype).


    Args:
        survey_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]
    """

    kwargs = _get_kwargs(
        survey_id=survey_id,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    survey_id: str,
    *,
    client: AuthenticatedClient,
) -> Optional[Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]]:
    r"""Retrieve Survey JSON Schema

     ## Introduction

    This documentation is meant to help explain the purpose and value of the Retrieve the JSON Schema
    endpoint in the public api. In short, this endpoint provides a [standard JSON Schema](https://json-
    schema.org/) that can be used for parsing the results from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint
    and the [Create Response Export](./legacyResponseImportsExports.json/paths/~1responseexports/post)
    endpoint when exporting in the json format. Finally this endpoint can also be useful for preparing
    responses for the [Create a New
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses/post) endpoint.

    # Example

    Note, before reading through this example, it might be helpful to take a quick detour and gain a bit
    of familiarity around Json Schemas.
    For this example we created a simple survey with a single multiple choice question `Which type of
    Ice Cream is your favorite?` and 4 options `Chocolate`, `Vanilla`, `Cookies & Cream`, and
    `Strawberry`.
    Here's an example response that we requested from the [Retrieve a Survey
    Response](./singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{responseId}/get) endpoint:

    ```json
    {
        \"responseId\": \"R_240YEnuWbbwZOy1\",
        \"values\": {
            \"startDate\": \"2020-05-13T19:26:15Z\",
            \"endDate\": \"2020-05-13T19:26:18Z\",
            \"status\": 0,
            \"ipAddress\": \"70.103.180.85\",
            \"progress\": 100,
            \"duration\": 2,
            \"finished\": 1,
            \"recordedDate\": \"2020-05-13T19:26:18.487Z\",
            \"locationLatitude\": \"40.3941955566\",
            \"locationLongitude\": \"-111.848297119\",
            \"distributionChannel\": \"anonymous\",
            \"userLanguage\": \"EN\",
            \"QID1\": 3,
            \"QID1_DO\": [
                \"1\",
                \"2\",
                \"3\",
                \"4\"
            ]
        },
        \"labels\": {
            \"status\": \"IP Address\",
            \"finished\": \"True\",
            \"QID1\": \"Cookies & Cream\",
            \"QID1_DO\": [
                \"Chocolate\",
                \"Vanilla\",
                \"Cookies & Cream\",
                \"Strawberry\"
            ]
        },
        \"displayedFields\": [
            \"QID1\"
        ],
        \"displayedValues\": {
            \"QID1\": [
                1,
                2,
                3,
                4
            ]
        }
    }
    ```

    Now, let's look at the pieces of the schema to understand them a little better. Note: this
    documentation won't attempt to cover all the aspects of the JSON Schema specification, for more
    information see [https://json-schema.org/](https://json-schema.org/).

    ```json
    {
        \"$schema\": \"https://json-schema.org/draft/2019-09/schema\",
        \"$id\": \"http://qualtrics.com/API/v3/surveys/SV_ai71yqu2ZLvpadf/response-schema\",
        \"title\": \"JSON Schema Example (SV_ai71yqu2ZLvpadf) - Survey Response\",
        \"description\": \"This is a schema for survey responses from SV_ai71yqu2ZLvpadf\",
        \"type\": \"object\",
    ```

    The schema starts with several standard fields from the Json Schema specification. The value for
    `$schema` is explained [here](https://json-schema.org/draft/2019-09/json-schema-
    core.html#rfc.section.8.1.1) and the value for `$id` is explained [here](https://json-
    schema.org/draft/2019-09/json-schema-core.html#id-keyword). Then, the schema is given a title and a
    description. Finally, the schema is defined for a JSON [object](https://json-
    schema.org/draft/2019-09/json-schema-core.html#rfc.section.4.2.1) primitive type.

    ```json
        \"required\": [
            \"values\",
            \"displayedFields\",
            \"displayedValues\",
            \"labels\",
            \"responseId\"
        ],
    ```

    Next, the `required` field specifies that our JSON `object` will have the certain required fields.
    Looking back at our example response, we see that each of these fields is present.

    ```json
        \"properties\": {
            \"responseId\": {
                \"type\": \"string\"
            },
    ```

    Inside of the `properties` object, each field previously mentioned in the `required` array are given
    a schema. The simplest of these fields is the responseId field which will just be provided as a
    string.

    ```json
            \"displayedFields\": {
                \"type\": \"array\",
                \"description\": \"A list of questions or question rows that were displayed to the
    survey respondent\",
                \"items\": {
                    \"type\": \"string\"
                }
            },
            \"displayedValues\": {
                \"type\": \"object\",
                \"description\": \"A list of the possible answers shown to the survey respondent for
    each question or question row\",
                \"additionalProperties\": {
                    \"type\": \"array\"
                }
            },
            \"labels\": {
                \"type\": \"object\",
                \"description\": \"The labels of the answers given by the survey respondent for each
    question or question row\",
                \"additionalProperties\": true
            },
    ```

    The next fields `displayedFields`, `displayedValues`, and `labels` have descriptions built in to the
    schema. Again, we can see how these schemas translate into a response object by consulting our
    example.

    ```json
            \"values\": {
                \"type\": \"object\",
                \"description\": \"The answers given by the survey respondent\",
                \"properties\": {
                    \"QID1\": {
                        \"description\": \"Which type of Ice Cream is your favorite?\",
                        \"dataType\": \"question\",
                        \"type\": \"number\",
                        \"oneOf\": [
                            {
                                \"label\": \"Chocolate\",
                                \"const\": 1
                            },
                            {
                                \"label\": \"Vanilla\",
                                \"const\": 2
                            },
                            {
                                \"label\": \"Cookies & Cream\",
                                \"const\": 3
                            },
                            {
                                \"label\": \"Strawberry\",
                                \"const\": 4
                            }
                        ],
                        \"exportTag\": \"Q1\",
                        \"questionId\": \"QID1\"
                    },
                    \"recordedDate\": {
                        \"description\": \"Recorded Date\",
                        \"dataType\": \"metadata\",
                        \"type\": \"string\",
                        \"format\": \"date-time\",
                        \"exportTag\": \"RecordedDate\"
                    },
                    \"duration\": {
                        \"description\": \"Duration (in seconds)\",
                        \"dataType\": \"metadata\",
                        \"type\": \"number\",
                        \"exportTag\": \"Duration (in seconds)\"
                    },
    ```

    Finally, the values property defines the schema for each value available in the response. I've
    pulled a couple of the fields from the schema into this documentation but we can see from the
    example response that there are many more fields in this schema. Here we can see a `description` for
    each field as well as a `type` and several other pieces of information. We can see that the `type`
    matches up with the JSON data type for each value as found in the example response. The `dataType`
    can be used to know the source for that value and the `exportTag` is useful for understanding the
    header columns when using the [Create Response
    Export](./responseImportsExports.json/paths/~1surveys~1{surveyId}~1export-responses/post) endpoint
    with the CSV or TSV formats. Lastly, oneOf enumerates the possible options for values that have a
    known, finite domain. For this example, a multiple choice question would have a oneOf field since
    there are only 4 possible values. However, a text entry field that only allows numbers would not
    have a oneOf field, since the domain can't be enumerated.

    ## Additional Tooling for JSON Schemas

    One of the benefits of having a schema is that it can be used to automate some of the parsing and
    usage of json schemas. One such tool for automating this step is
    [Quicktype](https://app.quicktype.io/#l=schema). More information about that tool can be found on
    [Github](https://github.com/quicktype/quicktype).


    Args:
        survey_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[DefaultErrorResponse, GetSchemaResponse, SurveyNotFoundResponse]
    """

    return (
        await asyncio_detailed(
            survey_id=survey_id,
            client=client,
        )
    ).parsed



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.create_response_response_result import CreateResponseResponseResult
    from ..models.meta import Meta


T = TypeVar("T", bound="CreateResponseResponse")


@_attrs_define
class CreateResponseResponse:
    """
    Attributes:
        result (CreateResponseResponseResult):
        meta (Meta):
    """

    result: "CreateResponseResponseResult"
    meta: "Meta"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        result = self.result.to_dict()

        meta = self.meta.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "result": result,
                "meta": meta,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.create_response_response_result import (
            CreateResponseResponseResult,
        )
        from ..models.meta import Meta

        d = src_dict.copy()
        result = CreateResponseResponseResult.from_dict(d.pop("result"))

        meta = Meta.from_dict(d.pop("meta"))

        create_response_response = cls(
            result=result,
            meta=meta,
        )

        create_response_response.additional_properties = d
        return create_response_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CreateResponseWithFileAttachmentsRequestFileMapping")


@_attrs_define
class CreateResponseWithFileAttachmentsRequestFileMapping:
    """
    Attributes:
        file1 (Union[Unset, str]): The QuestionID of the file upload question to map file1 to Example: QID1.
        file2 (Union[Unset, str]): The QuestionID of the file upload question to map file2 to Example: QID2.
        file3 (Union[Unset, str]): The QuestionID of the file upload question to map file3 to Example: QID3.
        file4 (Union[Unset, str]): The QuestionID of the file upload question to map file4 to Example: QID4.
        file5 (Union[Unset, str]): The QuestionID of the file upload question to map file5 to Example: QID5.
    """

    file1: Union[Unset, str] = UNSET
    file2: Union[Unset, str] = UNSET
    file3: Union[Unset, str] = UNSET
    file4: Union[Unset, str] = UNSET
    file5: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        file1 = self.file1
        file2 = self.file2
        file3 = self.file3
        file4 = self.file4
        file5 = self.file5

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if file1 is not UNSET:
            field_dict["file1"] = file1
        if file2 is not UNSET:
            field_dict["file2"] = file2
        if file3 is not UNSET:
            field_dict["file3"] = file3
        if file4 is not UNSET:
            field_dict["file4"] = file4
        if file5 is not UNSET:
            field_dict["file5"] = file5

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        file1 = d.pop("file1", UNSET)

        file2 = d.pop("file2", UNSET)

        file3 = d.pop("file3", UNSET)

        file4 = d.pop("file4", UNSET)

        file5 = d.pop("file5", UNSET)

        create_response_with_file_attachments_request_file_mapping = cls(
            file1=file1,
            file2=file2,
            file3=file3,
            file4=file4,
            file5=file5,
        )

        create_response_with_file_attachments_request_file_mapping.additional_properties = (
            d
        )
        return create_response_with_file_attachments_request_file_mapping

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="GetFiltersListResponseResultElementsItem")


@_attrs_define
class GetFiltersListResponseResultElementsItem:
    """
    Attributes:
        filter_name (str): The name of the filter.
        creation_date (datetime.datetime): The date and time the filter was created.
        filter_id (str): The ID of the filter.
    """

    filter_name: str
    creation_date: datetime.datetime
    filter_id: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        filter_name = self.filter_name
        creation_date = self.creation_date.isoformat()

        filter_id = self.filter_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "filterName": filter_name,
                "creationDate": creation_date,
                "filterId": filter_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        filter_name = d.pop("filterName")

        creation_date = isoparse(d.pop("creationDate"))

        filter_id = d.pop("filterId")

        get_filters_list_response_result_elements_item = cls(
            filter_name=filter_name,
            creation_date=creation_date,
            filter_id=filter_id,
        )

        get_filters_list_response_result_elements_item.additional_properties = d
        return get_filters_list_response_result_elements_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import re
import urllib.parse
from functools import cache
from typing import Any

import numpy as np
import pandas as pd
import sqlalchemy
from attr import asdict
from googleapiutils2 import Sheets
from loguru import logger
from sqlalchemy import (
    VARCHAR,
    Boolean,
    Column,
    DateTime,
    Float,
    Integer,
    MetaData,
    Table,
)
from sqlalchemy import Text as SQLAlchemyText

from qualtrics_utils.misc import T
from qualtrics_utils.surveys_response_import_export_api_client.types import UNSET

RE_HTML_TAG = re.compile(r"<(.|\s)*?>")

RE_SPACE = re.compile(r"&nbsp;")

RE_WHITESPACE = re.compile("\s+")


def normalize_whitespace(s: str) -> str:
    s = re.sub(RE_WHITESPACE, " ", s)
    return s.strip()


def normalize_html_string(s: str) -> str:
    s = re.sub(RE_HTML_TAG, "", s)
    s = re.sub(RE_SPACE, " ", s)
    s = normalize_whitespace(s)
    return s


def quote_value(value: str, quote: str = "`") -> str:
    if re.match(re.compile(quote + "(.*)" + quote), value) is not None:
        return value
    else:
        return f"{quote}{value}{quote}"


def get_url_params(url: str) -> dict[str, list[str]]:
    """Get the components of the given URL."""
    return urllib.parse.parse_qs(urllib.parse.urlparse(url).query)


def get_id_from_url(
    url: str, adjacent_options: list[str], query_param_options: list[str]
) -> str:
    """
    Extracts the ID from the URL provided.

    This function supports URLs of the form:
    - '.../x/{id}/...' for x in adjacent_options
    - With a query parameter named 'x': '...?x={x}...' for x in query_param_options

    We prioritize the adjacent_options over the query_param_options, and short circuit therein if we find a match.

    Args:
        url (str): The URL string from which to extract the ID.
    """
    paths = url.split("/")

    def get_adjacent(x: str) -> str | None:
        if x in paths and (t_ix := paths.index(x) + 1) < len(paths):
            return paths[t_ix]
        else:
            return None

    for i in adjacent_options:
        if (id := get_adjacent(i)) is not None:
            return id

    params = get_url_params(url)
    for i in query_param_options:
        if (ids := params.get(i)) is not None:
            return ids[0]

    raise ValueError(f"Could not parse file URL of {url}")


ADJACENT_OPTIONS = ["form", "survey-builder", "surveys"]
QUERY_PARAM_OPTIONS = ["surveyId", "id"]


@cache
def parse_file_id(
    file_id: str,
    adjacent_options: list[str] = ADJACENT_OPTIONS,
    query_param_options: list[str] = QUERY_PARAM_OPTIONS,
) -> str:
    """
    Parse the given file_id which could be an ID string, URL string or a dictionary object.

    This function supports the following formats:
    - Direct ID string: '123456789'
    - URL formats supported by 'get_id_from_url' function.
    - Dictionary object with 'id' or 'surveyId' as keys.

    Args:
        file_id (str): The ID string or URL or dictionary from which to extract the ID.
    """

    def parse(file_id: str) -> str:
        if "http" in file_id:
            return get_id_from_url(
                file_id,
                adjacent_options=adjacent_options,
                query_param_options=query_param_options,
            )
        else:
            return file_id

    def obj_to_id(file: str) -> str:
        if isinstance(file, str):
            return file
        elif isinstance(file, dict):
            for i in query_param_options:
                if (id := file.get(i)) is not None:
                    return id

        return None

    if (id := obj_to_id(file_id)) is not None:
        return parse(id)
    else:
        return file_id


def reset_request_defaults(request: T, set_items: dict[str, Any]) -> T:
    """Set all attributes of d to UNSET except for those in set_items
    This is useful for when you want all defaults to be UNSET, but you want to set some attributes to a value.

    Args:
        request (Any): Input request object from an OpenAPI client.
        set_items (dict[str, Any]): Dictionary of attributes that were explicitly set.
    """
    # request must have a to_dict() and from_dict() method
    if not hasattr(request, "to_dict") or not hasattr(request, "from_dict"):
        raise ValueError("request must have a to_dict() and from_dict() method")

    for k, v in asdict(request).items():  # type: ignore
        if k not in set_items and not isinstance(v, dict):
            setattr(request, k, UNSET)

    return request


def rename_columns(
    df: pd.DataFrame, codebook: list[dict[str, Any]], verbose: bool = True
) -> pd.DataFrame:
    renaming_map = {}
    for question in codebook:
        root_q_num = question["question_number"]
        sub_questions = question.get("questions")

        if root_q_num in df.columns:
            if verbose:
                renaming_map[
                    root_q_num
                ] = f"{root_q_num} - {question['question_string']}"
            else:
                renaming_map[root_q_num] = f"{root_q_num}"

        if sub_questions is not None and len(sub_questions) >= 1:
            for sub_question in sub_questions:
                q_num = sub_question["question_number"]
                if not q_num in df.columns:
                    continue

                if verbose:
                    renaming_map[
                        q_num
                    ] = f"{root_q_num} - {sub_question['question_string']}"
                else:
                    renaming_map[q_num] = f"{sub_question['question_number']}"

    df.rename(columns=renaming_map, inplace=True, errors="ignore")
    return df


def coalesce_multiselect(
    df: pd.DataFrame,
    codebook: list[dict[str, Any]],
    delimiter: str = ", ",
    use_multiple: bool = True,
) -> pd.DataFrame:
    """Coalesce a multi-select column into a single column.

    For example, if we have a multi-select question with the following options:
        - Option 1
        - Option 2
        - Option 3

    These will be typically split into individual columns, with each cell potentially being a non-null value.
    In many case, a multi-select question will have one response, and the rest will be null.

    This function will coalesce all the non-null values into a single column.
    If multiple values are present, then the values will be handled as follows:
        - If use_multiple is True, then the value will be set to "Multiple"
        - If use_multiple is False, then then value will be the concatenation of the values, separated by the `delimiter`.

    Args:
        - df (pd.DataFrame): The DataFrame to coalesce.
        - codebook (list[dict[str, Any]]): The codebook for the survey.
        - delimiter (str): The delimiter to use when joining multiple values.
        - use_multiple (bool): Whether to use the "Multiple" string when multiple values are present.
    """

    def join(x: pd.Series) -> str:
        l = x.dropna().tolist()

        if use_multiple and len(l) > 1:
            return "Multiple"
        else:
            return delimiter.join(l)

    root_questions = {}

    for question in codebook:
        if question["question_type"] == "MC":
            root_q_num = question["question_number"]
            sub_questions = question.get("questions")

            if sub_questions is None or len(sub_questions) <= 1:
                continue

            sub_question_numbers = [
                sub_question["question_number"] for sub_question in sub_questions
            ]
            sub_question_columns = [
                col for col in df.columns if col in sub_question_numbers
            ]
            root_questions[root_q_num] = df[sub_question_columns].apply(
                lambda x: join(x), axis=1
            )
            df.drop(columns=sub_question_columns, inplace=True)

    for root_q_num, root_question_df in root_questions.items():
        df[root_q_num] = root_question_df

    return df


def dtype_to_sqlalchemy(dtype: Any, index: bool = False):
    if isinstance(dtype, pd.Int64Dtype) or np.issubdtype(dtype, np.integer):
        return Integer
    elif isinstance(dtype, pd.Float64Dtype) or np.issubdtype(dtype, np.floating):
        return Float
    elif isinstance(dtype, pd.DatetimeTZDtype) or np.issubdtype(dtype, np.datetime64):
        return DateTime
    elif isinstance(dtype, pd.StringDtype) or (
        np.issubdtype(dtype, np.dtype("O"))
        or np.issubdtype(dtype, np.dtype("S"))
        or np.issubdtype(dtype, np.dtype("U"))
    ):
        if not index:
            return SQLAlchemyText
        else:
            return VARCHAR(255)
    elif isinstance(dtype, pd.BooleanDtype) or np.issubdtype(dtype, np.bool_):
        return Boolean
    else:
        raise ValueError(f"Unsupported dtype: {dtype}")


def generate_sql_schema(
    df: pd.DataFrame,
    table_name: str,
    index_as_pk: bool = False,
    auto_increment: bool = True,
):
    """Generate a SQLAlchemy table schema from a Pandas DataFrame.

    Args:
        df (pd.DataFrame): The DataFrame from which to generate the schema.
        table_name (str): The name of the table.
        index_as_pk (bool, optional): Whether to use the index as the primary key. Defaults to False.
        auto_increment (bool, optional): Whether to use autoincrement for the primary key. Defaults to True.
    """
    metadata = MetaData()
    columns: list[Column] = [
        Column(str(name), dtype_to_sqlalchemy(dtype))
        for name, dtype in df.dtypes.items()
    ]
    columns = [
        Column(
            str(name),
            dtype_to_sqlalchemy(dtype, index=index_as_pk),
            primary_key=index_as_pk,
        )
        for name, dtype in df.index.to_frame().dtypes.items()
    ] + columns

    if auto_increment:
        columns.insert(0, Column("id", Integer, primary_key=True, autoincrement=True))

    table = Table(table_name, metadata, *columns)
    return table


def qualtrics_schema_to_dtypes(
    schema: dict[str, Any],
    use_labels: bool = True,
    skip_embedded_data: bool = True,
    dtypes: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """
    Transforms a Qualtrics response schema into a map of column names to Pandas/NumPy data types,
    with an option to use string types for labeled questions.

    :param schema: Qualtrics response schema in dictionary format.
    :param use_labels: If True, map questions with labels to string type.
    :return: A dictionary mapping column names to Pandas/NumPy data types.
    """
    if dtypes is None:
        dtypes = {}

    t_types: dict[str, Any] = {}
    properties: dict = schema["result"]["properties"]["values"]["properties"]

    LABELED_TYPES = ["oneOf", "anyOf"]
    DISPLAY_ORDER = " - Display Order"

    for key, value in properties.items():
        name = value.get("exportTag")
        type = value.get("type")
        format_type = value.get("format")

        data_type = value.get("dataType")

        items = value.get("items", {})
        description = value.get("description", "")

        if skip_embedded_data and data_type == "embeddedData":
            continue

        # prefer format type over type
        if format_type is not None:
            type = format_type

        # skip over display order columns
        if description.endswith(DISPLAY_ORDER):
            continue

        is_labeled = any(
            [
                (label_type in value) or (label_type in items)
                for label_type in LABELED_TYPES
            ]
        )

        if is_labeled and use_labels:
            t_types[name] = str
            continue

        match type.lower():
            case "string":
                t_types[name] = str
            case "boolean":
                t_types[name] = bool
            case "date-time" | "date" | "time":
                t_types[name] = np.datetime64
            case "number" | "array":
                if is_labeled:
                    t_types[name] = pd.Int64Dtype()
                else:
                    t_types[name] = float
            case _:
                t_types[name] = str

    return {
        **t_types,
        **dtypes,
    }


def create_mysql_engine(
    username: str, password: str, host: str, port: str, database: str, **kwargs: Any
) -> sqlalchemy.engine.Engine:
    engine_str = f"mysql+pymysql://{username}:{password}@{host}:{port}/{database}"
    engine = sqlalchemy.create_engine(engine_str)
    return engine


def drop_if_exists(table_name: str, conn: sqlalchemy.Connection):
    metadata = sqlalchemy.MetaData()
    table = sqlalchemy.Table(table_name, metadata)

    if conn.dialect.has_table(conn, table_name):
        table.drop(conn)


def delete_sheet_if_exists(sheet_name: str, spreadsheet_id: str, sheets: Sheets):
    try:
        if sheets.has(name=sheet_name, spreadsheet_id=spreadsheet_id):
            sheets.delete(name=sheet_name, spreadsheet_id=spreadsheet_id)
    except Exception as e:
        logger.error(f"Error deleting sheet {sheet_name} from {spreadsheet_id}: {e}")
        pass

    return



""" Contains all the data models used in inputs/outputs """

from .create_import_json_body import CreateImportJsonBody
from .create_import_json_body_format import CreateImportJsonBodyFormat
from .creation_response import CreationResponse
from .creation_response_invalid import CreationResponseInvalid
from .creation_response_result import CreationResponseResult
from .creation_response_too_large import CreationResponseTooLarge
from .export_creation_request import ExportCreationRequest
from .export_creation_request_format import ExportCreationRequestFormat
from .export_status_response import ExportStatusResponse
from .export_status_response_result import ExportStatusResponseResult
from .get_filters_list_response import GetFiltersListResponse
from .get_filters_list_response_result import GetFiltersListResponseResult
from .get_filters_list_response_result_elements_item import (
    GetFiltersListResponseResultElementsItem,
)
from .import_by_url_creation_request import ImportByURLCreationRequest
from .import_status_response import ImportStatusResponse
from .import_status_response_not_found import ImportStatusResponseNotFound
from .import_status_response_result import ImportStatusResponseResult
from .internal_error import InternalError
from .meta import Meta
from .meta_error import MetaError
from .request_status import RequestStatus
from .survey_not_found import SurveyNotFound

__all__ = (
    "CreateImportJsonBody",
    "CreateImportJsonBodyFormat",
    "CreationResponse",
    "CreationResponseInvalid",
    "CreationResponseResult",
    "CreationResponseTooLarge",
    "ExportCreationRequest",
    "ExportCreationRequestFormat",
    "ExportStatusResponse",
    "ExportStatusResponseResult",
    "GetFiltersListResponse",
    "GetFiltersListResponseResult",
    "GetFiltersListResponseResultElementsItem",
    "ImportByURLCreationRequest",
    "ImportStatusResponse",
    "ImportStatusResponseNotFound",
    "ImportStatusResponseResult",
    "InternalError",
    "Meta",
    "MetaError",
    "RequestStatus",
    "SurveyNotFound",
)



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.create_response_request import CreateResponseRequest
from ...models.create_response_response import CreateResponseResponse
from ...models.create_response_with_file_attachments_request import (
    CreateResponseWithFileAttachmentsRequest,
)
from ...models.default_error_response import DefaultErrorResponse
from ...models.survey_not_found_response import SurveyNotFoundResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    survey_id: str,
    *,
    multipart_data: CreateResponseWithFileAttachmentsRequest,
    json_body: CreateResponseRequest,
    idempotency_key: Union[Unset, str] = UNSET,
) -> Dict[str, Any]:
    headers = {}
    if not isinstance(idempotency_key, Unset):
        headers["Idempotency-Key"] = idempotency_key

    json_body.to_dict()

    multipart_multipart_data = multipart_data.to_multipart()

    return {
        "method": "post",
        "url": "/surveys/{surveyId}/responses".format(
            surveyId=survey_id,
        ),
        "files": multipart_multipart_data,
        "headers": headers,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
]:
    if response.status_code == HTTPStatus.OK:
        response_200 = CreateResponseResponse.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = DefaultErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = DefaultErrorResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = SurveyNotFoundResponse.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = DefaultErrorResponse.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    multipart_data: CreateResponseWithFileAttachmentsRequest,
    json_body: CreateResponseRequest,
    idempotency_key: Union[Unset, str] = UNSET,
) -> Response[
    Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
]:
    r"""Create a New Response

     <!--From Readme-->
    <!-- theme: info -->

    >### Import Response Format
    >The fields in `values` depend on the survey definition. The easiest way to find this format is to
    perform a get a single response call for the survey which you are trying import values for [Retrieve
    a Survey Response](../reference/singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{respon
    seId}/get). The result of that call will demonstrate the format for the Create a New Response Call.


    <!--From Readme-->
    <!-- theme: info -->

    >### Uploading Files for File Upload Questions
    >Files can be included in your import to be associated with file upload questions in your survey.
    Select `multipart/form-data` as the request type in the documentation below to see the full request
    format.
    >
    >`multipart/form-data` requests allow you to send multiple different entities as body parts in a
    single request. We can use this to create a request to upload a survey response with file
    attachments. A minimum of two body parts are required in addition to the files you wish to upload.
    >1. `response`: This body part is the JSON representation of the survey response. As with the
    regular `application/json` POST request for this endpoint, the minimum required object is
    `{\"values\": {}}`. Note that you do not need to include any file metadata in the survey response as
    the necessary metadata will automatically be added to the response from the files that you upload
    with it.
    >1. `fileMapping`: This body part is also a JSON object. Its purpose is to tell the importer which
    files go with which questions in your survey. E.g. if we wanted to import a file with the body part
    key of `file1` and we wanted that file to be associated with the file upload question `QID1`, our
    `fileMapping` object would be `{\"file1\": \"QID1\"}`.
    >For the remainder of your request you can attach up to five files. Each file must have a body part
    key of `file1`, `file2`, `file3`, `file4`, or `file5`.


    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):
        multipart_data (CreateResponseWithFileAttachmentsRequest):
        json_body (CreateResponseRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]]
    """

    kwargs = _get_kwargs(
        survey_id=survey_id,
        multipart_data=multipart_data,
        json_body=json_body,
        idempotency_key=idempotency_key,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    multipart_data: CreateResponseWithFileAttachmentsRequest,
    json_body: CreateResponseRequest,
    idempotency_key: Union[Unset, str] = UNSET,
) -> Optional[
    Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
]:
    r"""Create a New Response

     <!--From Readme-->
    <!-- theme: info -->

    >### Import Response Format
    >The fields in `values` depend on the survey definition. The easiest way to find this format is to
    perform a get a single response call for the survey which you are trying import values for [Retrieve
    a Survey Response](../reference/singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{respon
    seId}/get). The result of that call will demonstrate the format for the Create a New Response Call.


    <!--From Readme-->
    <!-- theme: info -->

    >### Uploading Files for File Upload Questions
    >Files can be included in your import to be associated with file upload questions in your survey.
    Select `multipart/form-data` as the request type in the documentation below to see the full request
    format.
    >
    >`multipart/form-data` requests allow you to send multiple different entities as body parts in a
    single request. We can use this to create a request to upload a survey response with file
    attachments. A minimum of two body parts are required in addition to the files you wish to upload.
    >1. `response`: This body part is the JSON representation of the survey response. As with the
    regular `application/json` POST request for this endpoint, the minimum required object is
    `{\"values\": {}}`. Note that you do not need to include any file metadata in the survey response as
    the necessary metadata will automatically be added to the response from the files that you upload
    with it.
    >1. `fileMapping`: This body part is also a JSON object. Its purpose is to tell the importer which
    files go with which questions in your survey. E.g. if we wanted to import a file with the body part
    key of `file1` and we wanted that file to be associated with the file upload question `QID1`, our
    `fileMapping` object would be `{\"file1\": \"QID1\"}`.
    >For the remainder of your request you can attach up to five files. Each file must have a body part
    key of `file1`, `file2`, `file3`, `file4`, or `file5`.


    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):
        multipart_data (CreateResponseWithFileAttachmentsRequest):
        json_body (CreateResponseRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
    """

    return sync_detailed(
        survey_id=survey_id,
        client=client,
        multipart_data=multipart_data,
        json_body=json_body,
        idempotency_key=idempotency_key,
    ).parsed


async def asyncio_detailed(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    multipart_data: CreateResponseWithFileAttachmentsRequest,
    json_body: CreateResponseRequest,
    idempotency_key: Union[Unset, str] = UNSET,
) -> Response[
    Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
]:
    r"""Create a New Response

     <!--From Readme-->
    <!-- theme: info -->

    >### Import Response Format
    >The fields in `values` depend on the survey definition. The easiest way to find this format is to
    perform a get a single response call for the survey which you are trying import values for [Retrieve
    a Survey Response](../reference/singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{respon
    seId}/get). The result of that call will demonstrate the format for the Create a New Response Call.


    <!--From Readme-->
    <!-- theme: info -->

    >### Uploading Files for File Upload Questions
    >Files can be included in your import to be associated with file upload questions in your survey.
    Select `multipart/form-data` as the request type in the documentation below to see the full request
    format.
    >
    >`multipart/form-data` requests allow you to send multiple different entities as body parts in a
    single request. We can use this to create a request to upload a survey response with file
    attachments. A minimum of two body parts are required in addition to the files you wish to upload.
    >1. `response`: This body part is the JSON representation of the survey response. As with the
    regular `application/json` POST request for this endpoint, the minimum required object is
    `{\"values\": {}}`. Note that you do not need to include any file metadata in the survey response as
    the necessary metadata will automatically be added to the response from the files that you upload
    with it.
    >1. `fileMapping`: This body part is also a JSON object. Its purpose is to tell the importer which
    files go with which questions in your survey. E.g. if we wanted to import a file with the body part
    key of `file1` and we wanted that file to be associated with the file upload question `QID1`, our
    `fileMapping` object would be `{\"file1\": \"QID1\"}`.
    >For the remainder of your request you can attach up to five files. Each file must have a body part
    key of `file1`, `file2`, `file3`, `file4`, or `file5`.


    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):
        multipart_data (CreateResponseWithFileAttachmentsRequest):
        json_body (CreateResponseRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]]
    """

    kwargs = _get_kwargs(
        survey_id=survey_id,
        multipart_data=multipart_data,
        json_body=json_body,
        idempotency_key=idempotency_key,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    survey_id: str,
    *,
    client: AuthenticatedClient,
    multipart_data: CreateResponseWithFileAttachmentsRequest,
    json_body: CreateResponseRequest,
    idempotency_key: Union[Unset, str] = UNSET,
) -> Optional[
    Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
]:
    r"""Create a New Response

     <!--From Readme-->
    <!-- theme: info -->

    >### Import Response Format
    >The fields in `values` depend on the survey definition. The easiest way to find this format is to
    perform a get a single response call for the survey which you are trying import values for [Retrieve
    a Survey Response](../reference/singleResponses.json/paths/~1surveys~1{surveyId}~1responses~1{respon
    seId}/get). The result of that call will demonstrate the format for the Create a New Response Call.


    <!--From Readme-->
    <!-- theme: info -->

    >### Uploading Files for File Upload Questions
    >Files can be included in your import to be associated with file upload questions in your survey.
    Select `multipart/form-data` as the request type in the documentation below to see the full request
    format.
    >
    >`multipart/form-data` requests allow you to send multiple different entities as body parts in a
    single request. We can use this to create a request to upload a survey response with file
    attachments. A minimum of two body parts are required in addition to the files you wish to upload.
    >1. `response`: This body part is the JSON representation of the survey response. As with the
    regular `application/json` POST request for this endpoint, the minimum required object is
    `{\"values\": {}}`. Note that you do not need to include any file metadata in the survey response as
    the necessary metadata will automatically be added to the response from the files that you upload
    with it.
    >1. `fileMapping`: This body part is also a JSON object. Its purpose is to tell the importer which
    files go with which questions in your survey. E.g. if we wanted to import a file with the body part
    key of `file1` and we wanted that file to be associated with the file upload question `QID1`, our
    `fileMapping` object would be `{\"file1\": \"QID1\"}`.
    >For the remainder of your request you can attach up to five files. Each file must have a body part
    key of `file1`, `file2`, `file3`, `file4`, or `file5`.


    Args:
        survey_id (str):
        idempotency_key (Union[Unset, str]):
        multipart_data (CreateResponseWithFileAttachmentsRequest):
        json_body (CreateResponseRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[CreateResponseResponse, DefaultErrorResponse, SurveyNotFoundResponse]
    """

    return (
        await asyncio_detailed(
            survey_id=survey_id,
            client=client,
            multipart_data=multipart_data,
            json_body=json_body,
            idempotency_key=idempotency_key,
        )
    ).parsed



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.schema_property_items_properties_x import (
        SchemaPropertyItemsPropertiesX,
    )
    from ..models.schema_property_items_properties_y import (
        SchemaPropertyItemsPropertiesY,
    )


T = TypeVar("T", bound="SchemaPropertyItemsProperties")


@_attrs_define
class SchemaPropertyItemsProperties:
    """
    Attributes:
        x (Union[Unset, SchemaPropertyItemsPropertiesX]):
        y (Union[Unset, SchemaPropertyItemsPropertiesY]):
    """

    x: Union[Unset, "SchemaPropertyItemsPropertiesX"] = UNSET
    y: Union[Unset, "SchemaPropertyItemsPropertiesY"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        x: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.x, Unset):
            x = self.x.to_dict()

        y: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.y, Unset):
            y = self.y.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if x is not UNSET:
            field_dict["x"] = x
        if y is not UNSET:
            field_dict["y"] = y

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.schema_property_items_properties_x import (
            SchemaPropertyItemsPropertiesX,
        )
        from ..models.schema_property_items_properties_y import (
            SchemaPropertyItemsPropertiesY,
        )

        d = src_dict.copy()
        _x = d.pop("x", UNSET)
        x: Union[Unset, SchemaPropertyItemsPropertiesX]
        if isinstance(_x, Unset):
            x = UNSET
        else:
            x = SchemaPropertyItemsPropertiesX.from_dict(_x)

        _y = d.pop("y", UNSET)
        y: Union[Unset, SchemaPropertyItemsPropertiesY]
        if isinstance(_y, Unset):
            y = UNSET
        else:
            y = SchemaPropertyItemsPropertiesY.from_dict(_y)

        schema_property_items_properties = cls(
            x=x,
            y=y,
        )

        schema_property_items_properties.additional_properties = d
        return schema_property_items_properties

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.create_import_json_body_format import CreateImportJsonBodyFormat
from ..types import UNSET, Unset

T = TypeVar("T", bound="CreateImportJsonBody")


@_attrs_define
class CreateImportJsonBody:
    """
    Attributes:
        format_ (CreateImportJsonBodyFormat): Allowed values are CSV or TSV. See guide before for instructions on how to
            format CSV or TSV files. Default: CreateImportJsonBodyFormat.CSV.
        file_url (str): The URL of a CSV or TSV file containing the survey responses to import. Example:
            https://mydomain.com/importfile.csv.
        charset (Union[Unset, str]):  Default: 'UTF-8'.
    """

    file_url: str
    format_: CreateImportJsonBodyFormat = CreateImportJsonBodyFormat.CSV
    charset: Union[Unset, str] = "UTF-8"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        format_ = self.format_.value

        file_url = self.file_url
        charset = self.charset

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "format": format_,
                "fileUrl": file_url,
            }
        )
        if charset is not UNSET:
            field_dict["charset"] = charset

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        format_ = CreateImportJsonBodyFormat(d.pop("format"))

        file_url = d.pop("fileUrl")

        charset = d.pop("charset", UNSET)

        create_import_json_body = cls(
            format_=format_,
            file_url=file_url,
            charset=charset,
        )

        create_import_json_body.additional_properties = d
        return create_import_json_body

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
